# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os
import math
import warnings

import galsim
from galsim_test_helpers import *
from galsim.utilities import single_threaded


#
# Note: all tests below were generated using the python interface to the RNG.  Eventually need tests
# for comparison against the C++!
#

precision = 10
# decimal point at which agreement is required for all double precision tests

precisionD = precision
precisionF = 5  # precision=10 does not make sense at single precision
precisionS = 1  # "precision" also a silly concept for ints, but allows all 4 tests to run in one go
precisionI = 1

# The number of values to generate when checking the mean and variance calculations.
# This is currently low enough to not dominate the time of the unit tests, but when changing
# something, it may be useful to add a couple zeros while testing.
nvals = 100000

testseed = 1000 # seed used for UniformDeviate for all tests
# Warning! If you change testseed, then all of the *Result variables below must change as well.

# the right answer for the first three uniform deviates produced from testseed
uResult = (0.11860922840423882, 0.21456799632869661, 0.43088198406621814)

# mean, sigma to use for Gaussian tests
gMean = 4.7
gSigma = 3.2
# the right answer for the first three Gaussian deviates produced from testseed
gResult = (6.3344979808161215, 6.2082355273987861, -0.069894693358302007)

# N, p to use for binomial tests
bN = 10
bp = 0.7
# the right answer for the first three binomial deviates produced from testseed
bResult = (9, 8, 7)

# mean to use for Poisson tests
pMean = 7
# the right answer for the first three Poisson deviates produced from testseed
pResult = (4, 5, 6)

# a & b to use for Weibull tests
wA = 4.
wB = 9.
# Tabulated results for Weibull
wResult = (5.3648053017485591, 6.3093033550873878, 7.7982696798921074)

# k & theta to use for Gamma tests
gammaK = 1.5
gammaTheta = 4.5
# Tabulated results for Gamma
gammaResult = (4.7375613139927157, 15.272973580418618, 21.485016362839747)

# n to use for Chi2 tests
chi2N = 30
# Tabulated results for Chi2
chi2Result = (32.209933900954049, 50.040002656028513, 24.301442486313896)

#function and min&max to use for DistDeviate function call tests
dmin=0.0
dmax=2.0
def dfunction(x):
    return x*x
# Tabulated results for DistDeviate function call
dFunctionResult = (0.9826461346196363, 1.1973307331701328, 1.5105900949284945)

# x and p arrays and interpolant to use for DistDeviate LookupTable tests
dx=[0.0, 1.0, 1.000000001, 2.999999999, 3.0, 4.0]
dp=[0.1, 0.1, 0.0    , 0.0    , 0.1, 0.1]
dLookupTable=galsim.LookupTable(x=dx,f=dp,interpolant='linear')
# Tabulated results for DistDeviate LookupTable call
dLookupTableResult = (0.23721845680847731, 0.42913599265739233, 0.86176396813243539)
# File with the same values
dLookupTableFile = os.path.join('random_data','dLookupTable.dat')


@timer
def test_uniform():
    """Test uniform random number generator
    """
    u = galsim.UniformDeviate(testseed)
    u2 = u.duplicate()
    u3 = galsim.UniformDeviate(u.serialize())
    testResult = (u(), u(), u())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(uResult), precision,
            err_msg='Wrong uniform random number sequence generated')
    testResult = (u2(), u2(), u2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(uResult), precision,
            err_msg='Wrong uniform random number sequence generated with duplicate')
    testResult = (u3(), u3(), u3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(uResult), precision,
            err_msg='Wrong uniform random number sequence generated from serialize')

    # Check that the mean and variance come out right
    u = galsim.UniformDeviate(testseed)
    vals = [u() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = 1./2.
    v = 1./12.
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from UniformDeviate')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from UniformDeviate')

    # Check discard
    u2 = galsim.UniformDeviate(testseed)
    u2.discard(nvals)
    v1, v2 = u(), u2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    assert u.has_reliable_discard
    assert not u.generates_in_pairs

    # Check seed, reset
    u.seed(testseed)
    testResult2 = (u(), u(), u())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong uniform random number sequence generated after seed')

    u.reset(testseed)
    testResult2 = (u(), u(), u())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong uniform random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    u.reset(rng)
    testResult2 = (u(), u(), u())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong uniform random number sequence generated after reset(rng)')

    # Check raw
    u2.reset(testseed)
    u2.discard(3)
    np.testing.assert_equal(u.raw(), u2.raw(),
            err_msg='Uniform deviates generate different raw values')

    rng2 = galsim.BaseDeviate(testseed)
    rng2.discard(4)
    np.testing.assert_equal(rng.raw(), rng2.raw(),
            err_msg='BaseDeviates generate different raw values after discard')

    # Check that two connected uniform deviates work correctly together.
    u2 = galsim.UniformDeviate(testseed)
    u.reset(u2)
    testResult2 = (u(), u2(), u())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong uniform random number sequence generated using two uds')
    u.seed(testseed)
    testResult2 = (u2(), u(), u2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong uniform random number sequence generated using two uds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    u.seed()
    testResult2 = (u(), u(), u())
    assert testResult2 != testResult
    u.reset()
    testResult3 = (u(), u(), u())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    u.reset()
    testResult4 = (u(), u(), u())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    u = galsim.UniformDeviate()
    testResult5 = (u(), u(), u())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Test generate
    u.seed(testseed)
    test_array = np.empty(3)
    u.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(uResult), precision,
            err_msg='Wrong uniform random number sequence from generate.')

    # Test add_generate
    u.seed(testseed)
    u.add_generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, 2.*np.array(uResult), precision,
            err_msg='Wrong uniform random number sequence from generate.')

    # Test generate with a float32 array
    u.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    u.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(uResult), precisionF,
            err_msg='Wrong uniform random number sequence from generate.')

    # Test add_generate
    u.seed(testseed)
    u.add_generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, 2.*np.array(uResult), precisionF,
            err_msg='Wrong uniform random number sequence from generate.')

    # Check that generated values are independent of number of threads.
    u1 = galsim.UniformDeviate(testseed)
    u2 = galsim.UniformDeviate(testseed)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        u1.generate(v1)
    with single_threaded(num_threads=10):
        u2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        u1.add_generate(v1)
    with single_threaded(num_threads=10):
        u2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(u, lambda x: x.serialize(), random=True)
    check_pickle(u, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(u, random=True)
    check_pickle(rng, random=True)
    assert 'UniformDeviate' in repr(u)
    assert 'UniformDeviate' in str(u)
    assert isinstance(eval(repr(u)), galsim.UniformDeviate)
    assert isinstance(eval(str(u)), galsim.UniformDeviate)
    assert isinstance(eval(repr(rng)), galsim.BaseDeviate)
    assert isinstance(eval(str(rng)), galsim.BaseDeviate)

    # Check that we can construct a UniformDeviate from None, and that it depends on dev/random.
    u1 = galsim.UniformDeviate(None)
    u2 = galsim.UniformDeviate(None)
    assert u1 != u2, "Consecutive UniformDeviate(None) compared equal!"
    # We shouldn't be able to construct a UniformDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.UniformDeviate, dict())
    assert_raises(TypeError, galsim.UniformDeviate, list())
    assert_raises(TypeError, galsim.UniformDeviate, set())

    assert_raises(TypeError, u.seed, '123')
    assert_raises(TypeError, u.seed, 12.3)


@timer
def test_gaussian():
    """Test Gaussian random number generator
    """
    g = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    g2 = g.duplicate()
    g3 = galsim.GaussianDeviate(g.serialize(), mean=gMean, sigma=gSigma)
    testResult = (g(), g(), g())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gResult), precision,
            err_msg='Wrong Gaussian random number sequence generated')
    testResult = (g2(), g2(), g2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gResult), precision,
            err_msg='Wrong Gaussian random number sequence generated with duplicate')
    testResult = (g3(), g3(), g3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gResult), precision,
            err_msg='Wrong Gaussian random number sequence generated from serialize')

    # Check that the mean and variance come out right
    g = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    vals = [g() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = gMean
    v = gSigma**2
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from GaussianDeviate')
    np.testing.assert_almost_equal(var, v, 0,
            err_msg='Wrong variance from GaussianDeviate')

    # Check discard
    g2 = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    g2.discard(nvals)
    v1,v2 = g(),g2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    # Note: For Gaussian, this only works if nvals is even.
    g2 = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    g2.discard(nvals+1, suppress_warnings=True)
    v1,v2 = g(),g2()
    print('after %d vals, next one is %s, %s'%(nvals+1,v1,v2))
    assert v1 != v2
    assert g.has_reliable_discard
    assert g.generates_in_pairs

    # If don't explicitly suppress the warning, then a warning is emitted when n is odd.
    g2 = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    with assert_warns(galsim.GalSimWarning):
        g2.discard(nvals+1)

    # Check seed, reset
    g.seed(testseed)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated after seed')

    g.reset(testseed)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    g.reset(rng)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    g.reset(ud)
    testResult = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated after reset(ud)')

    # Check that two connected Gaussian deviates work correctly together.
    g2 = galsim.GaussianDeviate(testseed, mean=gMean, sigma=gSigma)
    g.reset(g2)
    # Note: GaussianDeviate generates two values at a time, so we have to compare them in pairs.
    testResult2 = (g(), g(), g2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated using two gds')
    g.seed(testseed)
    # For the same reason, after seeding one, we need to manually clear the other's cache:
    g2.clearCache()
    testResult2 = (g2(), g2(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Gaussian random number sequence generated using two gds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    g.seed()
    testResult2 = (g(), g(), g())
    assert testResult2 != testResult
    g.reset()
    testResult3 = (g(), g(), g())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    g.reset()
    testResult4 = (g(), g(), g())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    g = galsim.GaussianDeviate(mean=gMean, sigma=gSigma)
    testResult5 = (g(), g(), g())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Test generate
    g.seed(testseed)
    test_array = np.empty(3)
    g.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gResult), precision,
            err_msg='Wrong Gaussian random number sequence from generate.')

    # Test generate_from_variance.
    g2 = galsim.GaussianDeviate(testseed, mean=5, sigma=0.3)
    g3 = galsim.GaussianDeviate(testseed, mean=5, sigma=0.3)
    test_array.fill(gSigma**2)
    g2.generate_from_variance(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gResult)-gMean, precision,
            err_msg='Wrong Gaussian random number sequence from generate_from_variance.')
    # After running generate_from_variance, it should be back to using the specified mean, sigma.
    # Note: need to round up to even number for discard, since gd generates 2 at a time.
    g3.discard((len(test_array)+1)//2 * 2)
    print('g2,g3 = ',g2(),g3())
    assert g2() == g3()

    # Test generate with a float32 array.
    g.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    g.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gResult), precisionF,
            err_msg='Wrong Gaussian random number sequence from generate.')

    # Test generate_from_variance.
    g2.seed(testseed)
    test_array.fill(gSigma**2)
    g2.generate_from_variance(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gResult)-gMean, precisionF,
            err_msg='Wrong Gaussian random number sequence from generate_from_variance.')

    # Check that generated values are independent of number of threads.
    g1 = galsim.GaussianDeviate(testseed, mean=53, sigma=1.3)
    g2 = galsim.GaussianDeviate(testseed, mean=53, sigma=1.3)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        g1.generate(v1)
    with single_threaded(num_threads=10):
        g2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        g1.add_generate(v1)
    with single_threaded(num_threads=10):
        g2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)
    ud = galsim.UniformDeviate(testseed + 3)
    ud.generate(v1)
    v1 += 6.7
    v2[:] = v1
    with single_threaded():
        g1.generate_from_variance(v1)
    with single_threaded(num_threads=10):
        g2.generate_from_variance(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(g, lambda x: (x.serialize(), x.mean, x.sigma), random=True)
    check_pickle(g, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(g, random=True)
    assert 'GaussianDeviate' in repr(g)
    assert 'GaussianDeviate' in str(g)
    assert isinstance(eval(repr(g)), galsim.GaussianDeviate)
    assert isinstance(eval(str(g)), galsim.GaussianDeviate)

    # Check that we can construct a GaussianDeviate from None, and that it depends on dev/random.
    g1 = galsim.GaussianDeviate(None)
    g2 = galsim.GaussianDeviate(None)
    assert g1 != g2, "Consecutive GaussianDeviate(None) compared equal!"
    # We shouldn't be able to construct a GaussianDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.GaussianDeviate, dict())
    assert_raises(TypeError, galsim.GaussianDeviate, list())
    assert_raises(TypeError, galsim.GaussianDeviate, set())

    assert_raises(ValueError, galsim.GaussianDeviate, testseed, mean=1, sigma=-1)


@timer
def test_binomial():
    """Test binomial random number generator
    """
    b = galsim.BinomialDeviate(testseed, N=bN, p=bp)
    b2 = b.duplicate()
    b3 = galsim.BinomialDeviate(b.serialize(), N=bN, p=bp)
    testResult = (b(), b(), b())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(bResult), precision,
            err_msg='Wrong binomial random number sequence generated')
    testResult = (b2(), b2(), b2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(bResult), precision,
            err_msg='Wrong binomial random number sequence generated with duplicate')
    testResult = (b3(), b3(), b3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(bResult), precision,
            err_msg='Wrong binomial random number sequence generated from serialize')

    # Check that the mean and variance come out right
    b = galsim.BinomialDeviate(testseed, N=bN, p=bp)
    vals = [b() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = bN*bp
    v = bN*bp*(1.-bp)
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from BinomialDeviate')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from BinomialDeviate')

    # Check discard
    b2 = galsim.BinomialDeviate(testseed, N=bN, p=bp)
    b2.discard(nvals)
    v1,v2 = b(),b2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    assert b.has_reliable_discard
    assert not b.generates_in_pairs

    # Check seed, reset
    b.seed(testseed)
    testResult2 = (b(), b(), b())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated after seed')

    b.reset(testseed)
    testResult2 = (b(), b(), b())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    b.reset(rng)
    testResult2 = (b(), b(), b())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    b.reset(ud)
    testResult = (b(), b(), b())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated after reset(ud)')

    # Check that two connected binomial deviates work correctly together.
    b2 = galsim.BinomialDeviate(testseed, N=bN, p=bp)
    b.reset(b2)
    testResult2 = (b(), b2(), b())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated using two bds')
    b.seed(testseed)
    testResult2 = (b2(), b(), b2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong binomial random number sequence generated using two bds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.  However, in this case, there are few enough options
    # for the output that occasionally two of these match.  So we don't do the normal
    # testResult2 != testResult, etc.
    b.seed()
    testResult2 = (b(), b(), b())
    #assert testResult2 != testResult
    b.reset()
    testResult3 = (b(), b(), b())
    #assert testResult3 != testResult
    #assert testResult3 != testResult2
    b.reset()
    testResult4 = (b(), b(), b())
    #assert testResult4 != testResult
    #assert testResult4 != testResult2
    #assert testResult4 != testResult3
    b = galsim.BinomialDeviate(N=bN, p=bp)
    testResult5 = (b(), b(), b())
    #assert testResult5 != testResult
    #assert testResult5 != testResult2
    #assert testResult5 != testResult3
    #assert testResult5 != testResult4

    # Test generate
    b.seed(testseed)
    test_array = np.empty(3)
    b.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(bResult), precision,
            err_msg='Wrong binomial random number sequence from generate.')

    # Test generate with an int array
    b.seed(testseed)
    test_array = np.empty(3, dtype=int)
    b.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(bResult), precisionI,
            err_msg='Wrong binomial random number sequence from generate.')

    # Check that generated values are independent of number of threads.
    b1 = galsim.BinomialDeviate(testseed, N=17, p=0.7)
    b2 = galsim.BinomialDeviate(testseed, N=17, p=0.7)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        b1.generate(v1)
    with single_threaded(num_threads=10):
        b2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        b1.add_generate(v1)
    with single_threaded(num_threads=10):
        b2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(b, lambda x: (x.serialize(), x.n, x.p), random=True)
    check_pickle(b, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(b, random=True)
    assert 'BinomialDeviate' in repr(b)
    assert 'BinomialDeviate' in str(b)
    assert isinstance(eval(repr(b)), galsim.BinomialDeviate)
    assert isinstance(eval(str(b)), galsim.BinomialDeviate)

    # Check that we can construct a BinomialDeviate from None, and that it depends on dev/random.
    b1 = galsim.BinomialDeviate(None)
    b2 = galsim.BinomialDeviate(None)
    assert b1 != b2, "Consecutive BinomialDeviate(None) compared equal!"
    # We shouldn't be able to construct a BinomialDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.BinomialDeviate, dict())
    assert_raises(TypeError, galsim.BinomialDeviate, list())
    assert_raises(TypeError, galsim.BinomialDeviate, set())


@timer
def test_poisson():
    """Test Poisson random number generator
    """
    p = galsim.PoissonDeviate(testseed, mean=pMean)
    p2 = p.duplicate()
    p3 = galsim.PoissonDeviate(p.serialize(), mean=pMean)
    testResult = (p(), p(), p())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(pResult), precision,
            err_msg='Wrong Poisson random number sequence generated')
    testResult = (p2(), p2(), p2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(pResult), precision,
            err_msg='Wrong Poisson random number sequence generated with duplicate')
    testResult = (p3(), p3(), p3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(pResult), precision,
            err_msg='Wrong Poisson random number sequence generated from serialize')

    # Check that the mean and variance come out right
    p = galsim.PoissonDeviate(testseed, mean=pMean)
    vals = [p() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = pMean
    v = pMean
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from PoissonDeviate')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from PoissonDeviate')

    # Check discard
    p2 = galsim.PoissonDeviate(testseed, mean=pMean)
    p2.discard(nvals, suppress_warnings=True)
    v1,v2 = p(),p2()
    print('With mean = %d, after %d vals, next one is %s, %s'%(pMean,nvals,v1,v2))
    assert v1 == v2

    # With a very small mean value, Poisson reliably only uses 1 rng per value.
    # But at only slightly larger means, it sometimes uses two rngs for a single value.
    # Basically anything >= 10 causes this next test to have v1 != v2
    high_mean = 10
    p = galsim.PoissonDeviate(testseed, mean=high_mean)
    p2 = galsim.PoissonDeviate(testseed, mean=high_mean)
    vals = [p() for i in range(nvals)]
    p2.discard(nvals, suppress_warnings=True)
    v1,v2 = p(),p2()
    print('With mean = %d, after %d vals, next one is %s, %s'%(high_mean,nvals,v1,v2))
    assert v1 != v2
    assert not p.has_reliable_discard
    assert not p.generates_in_pairs

    # Discard normally emits a warning for Poisson
    p2 = galsim.PoissonDeviate(testseed, mean=pMean)
    with assert_warns(galsim.GalSimWarning):
        p2.discard(nvals)

    # Check seed, reset
    p = galsim.PoissonDeviate(testseed, mean=pMean)
    p.seed(testseed)
    testResult2 = (p(), p(), p())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated after seed')

    p.reset(testseed)
    testResult2 = (p(), p(), p())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    p.reset(rng)
    testResult2 = (p(), p(), p())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    p.reset(ud)
    testResult = (p(), p(), p())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated after reset(ud)')

    # Check that two connected poisson deviates work correctly together.
    p2 = galsim.PoissonDeviate(testseed, mean=pMean)
    p.reset(p2)
    testResult2 = (p(), p2(), p())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated using two pds')
    p.seed(testseed)
    testResult2 = (p2(), p(), p2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong poisson random number sequence generated using two pds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.  However, in this case, there are few enough options
    # for the output that occasionally two of these match.  So we don't do the normal
    # testResult2 != testResult, etc.
    p.seed()
    testResult2 = (p(), p(), p())
    #assert testResult2 != testResult
    p.reset()
    testResult3 = (p(), p(), p())
    #assert testResult3 != testResult
    #assert testResult3 != testResult2
    p.reset()
    testResult4 = (p(), p(), p())
    #assert testResult4 != testResult
    #assert testResult4 != testResult2
    #assert testResult4 != testResult3
    p = galsim.PoissonDeviate(mean=pMean)
    testResult5 = (p(), p(), p())
    #assert testResult5 != testResult
    #assert testResult5 != testResult2
    #assert testResult5 != testResult3
    #assert testResult5 != testResult4

    # Test generate
    p.seed(testseed)
    test_array = np.empty(3)
    p.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(pResult), precision,
            err_msg='Wrong poisson random number sequence from generate.')

    # Test generate with an int array
    p.seed(testseed)
    test_array = np.empty(3, dtype=int)
    p.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(pResult), precisionI,
            err_msg='Wrong poisson random number sequence from generate.')

    # Test generate_from_expectation
    p2 = galsim.PoissonDeviate(testseed, mean=77)
    test_array = np.array([pMean]*3, dtype=int)
    p2.generate_from_expectation(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(pResult), precisionI,
            err_msg='Wrong poisson random number sequence from generate_from_expectation.')
    # After generating, it should be back to mean=77
    test_array2 = np.array([p2() for i in range(100)])
    print('test_array2 = ',test_array2)
    print('mean = ',test_array2.mean())
    assert np.isclose(test_array2.mean(), 77, atol=2)

    # Check that generated values are independent of number of threads.
    # This should be trivial, since Poisson disables multi-threading, but check anyway.
    p1 = galsim.PoissonDeviate(testseed, mean=77)
    p2 = galsim.PoissonDeviate(testseed, mean=77)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        p1.generate(v1)
    with single_threaded(num_threads=10):
        p2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        p1.add_generate(v1)
    with single_threaded(num_threads=10):
        p2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(p, lambda x: (x.serialize(), x.mean), random=True)
    check_pickle(p, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(p, random=True)
    assert 'PoissonDeviate' in repr(p)
    assert 'PoissonDeviate' in str(p)
    assert isinstance(eval(repr(p)), galsim.PoissonDeviate)
    assert isinstance(eval(str(p)), galsim.PoissonDeviate)

    # Check that we can construct a PoissonDeviate from None, and that it depends on dev/random.
    p1 = galsim.PoissonDeviate(None)
    p2 = galsim.PoissonDeviate(None)
    assert p1 != p2, "Consecutive PoissonDeviate(None) compared equal!"
    # We shouldn't be able to construct a PoissonDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.PoissonDeviate, dict())
    assert_raises(TypeError, galsim.PoissonDeviate, list())
    assert_raises(TypeError, galsim.PoissonDeviate, set())


@timer
def test_poisson_highmean(run_slow):
    """Test Poisson random number generator with high (>2^30) mean (cf. Issue #881)

    It turns out that the boost poisson deviate class that we use maxes out at 2^31 and wraps
    around to -2^31.  We have code to automatically switch over to using a Gaussian deviate
    instead if the mean > 2^30 (factor of 2 from the problem to be safe).  Check that this
    works properly.
    """
    mean_vals =[ 2**30 + 50,  # Uses Gaussian
                 2**30 - 50,  # Uses Poisson
                 2**30,       # Uses Poisson (highest value of mean that does)
                 2**31,       # This is where problems happen if not using Gaussian
                 5.e20,       # Definitely would have problems with normal implementation.
               ]

    if run_slow:
        nvals = 10000000
        rtol_var = 1.e-3
    else:
        nvals = 100000
        rtol_var = 1.e-2

    for mean in mean_vals:
        print('Test PoissonDeviate with mean = ',mean)
        p = galsim.PoissonDeviate(testseed, mean=mean)
        p2 = p.duplicate()
        p3 = galsim.PoissonDeviate(p.serialize(), mean=mean)
        testResult = (p(), p(), p())
        testResult2 = (p2(), p2(), p2())
        testResult3 = (p3(), p3(), p3())
        np.testing.assert_allclose(
                testResult2, testResult, rtol=1.e-8,
                err_msg='PoissonDeviate.duplicate not equivalent for mean=%s'%mean)
        np.testing.assert_allclose(
                testResult3, testResult, rtol=1.e-8,
                err_msg='PoissonDeviate from serialize not equivalent for mean=%s'%mean)

        # Check that the mean and variance come out right
        p = galsim.PoissonDeviate(testseed, mean=mean)
        vals = [p() for i in range(nvals)]
        mu = np.mean(vals)
        var = np.var(vals)
        print('mean = ',mu,'  true mean = ',mean)
        print('var = ',var,'   true var = ',mean)
        np.testing.assert_allclose(mu, mean, rtol=1.e-5,
                err_msg='Wrong mean from PoissonDeviate with mean=%s'%mean)
        np.testing.assert_allclose(var, mean, rtol=rtol_var,
                err_msg='Wrong variance from PoissonDeviate with mean=%s'%mean)

        # Check discard
        p2 = galsim.PoissonDeviate(testseed, mean=mean)
        p2.discard(nvals, suppress_warnings=True)
        v1,v2 = p(),p2()
        print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
        if mean > 2**30:
            # Poisson doesn't have a reliable rng count (unless the mean is vv small).
            # But above 2**30 we're back to Gaussian, which is reliable.
            assert v1 == v2

        # Check that two connected poisson deviates work correctly together.
        p2 = galsim.PoissonDeviate(testseed, mean=mean)
        p.reset(p2)
        testResult2 = (p(), p(), p2())
        np.testing.assert_array_equal(
                testResult2, testResult,
                err_msg='Wrong poisson random number sequence generated using two pds')
        p.seed(testseed)
        p2.clearCache()
        testResult2 = (p2(), p2(), p())
        np.testing.assert_array_equal(
                testResult2, testResult,
                err_msg='Wrong poisson random number sequence generated using two pds after seed')

        # Test filling an image
        p.seed(testseed)
        testimage = galsim.ImageD(np.zeros((3, 1)))
        testimage.addNoise(galsim.DeviateNoise(p))
        np.testing.assert_array_equal(
                testimage.array.flatten(), testResult,
                err_msg='Wrong poisson random number sequence generated when applied to image.')

        # The PoissonNoise version also subtracts off the mean value
        rng = galsim.BaseDeviate(testseed)
        pn = galsim.PoissonNoise(rng, sky_level=mean)
        testimage.fill(0)
        testimage.addNoise(pn)
        np.testing.assert_array_equal(
                testimage.array.flatten(), np.array(testResult)-mean,
                err_msg='Wrong poisson random number sequence generated using PoissonNoise')

        # Check PoissonNoise variance:
        np.testing.assert_allclose(
                pn.getVariance(), mean, rtol=1.e-8,
                err_msg="PoissonNoise getVariance returns wrong variance")
        np.testing.assert_allclose(
                pn.sky_level, mean, rtol=1.e-8,
                err_msg="PoissonNoise sky_level returns wrong value")

        # Check that the noise model really does produce this variance.
        big_im = galsim.Image(2048,2048,dtype=float)
        big_im.addNoise(pn)
        var = np.var(big_im.array)
        print('variance = ',var)
        print('getVar = ',pn.getVariance())
        np.testing.assert_allclose(
                var, pn.getVariance(), rtol=rtol_var,
                err_msg='Realized variance for PoissonNoise did not match getVariance()')


@timer
def test_poisson_zeromean():
    """Make sure Poisson Deviate behaves sensibly when mean=0.
    """
    p = galsim.PoissonDeviate(testseed, mean=0)
    p2 = p.duplicate()
    p3 = galsim.PoissonDeviate(p.serialize(), mean=0)
    check_pickle(p, random=True)

    # Test direct draws
    testResult = (p(), p(), p())
    testResult2 = (p2(), p2(), p2())
    testResult3 = (p3(), p3(), p3())
    np.testing.assert_array_equal(testResult, 0)
    np.testing.assert_array_equal(testResult2, 0)
    np.testing.assert_array_equal(testResult3, 0)

    # Test generate
    test_array = np.empty(3, dtype=int)
    p.generate(test_array)
    np.testing.assert_array_equal(test_array, 0)
    p2.generate(test_array)
    np.testing.assert_array_equal(test_array, 0)
    p3.generate(test_array)
    np.testing.assert_array_equal(test_array, 0)

    # Test generate_from_expectation
    test_array = np.array([0,0,0])
    np.testing.assert_allclose(test_array, 0)
    test_array = np.array([1,0,4])
    assert test_array[0] != 0
    assert test_array[1] == 0
    assert test_array[2] != 0

    # Error raised if mean<0
    with assert_raises(ValueError):
        p = galsim.PoissonDeviate(testseed, mean=-0.1)
    with assert_raises(ValueError):
        p = galsim.PoissonDeviate(testseed, mean=-10)
    test_array = np.array([-1,1,4])
    with assert_raises(ValueError):
        p.generate_from_expectation(test_array)
    test_array = np.array([1,-1,-4])
    with assert_raises(ValueError):
        p.generate_from_expectation(test_array)

@timer
def test_weibull():
    """Test Weibull random number generator
    """
    w = galsim.WeibullDeviate(testseed, a=wA, b=wB)
    w2 = w.duplicate()
    w3 = galsim.WeibullDeviate(w.serialize(), a=wA, b=wB)
    testResult = (w(), w(), w())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(wResult), precision,
            err_msg='Wrong Weibull random number sequence generated')
    testResult = (w2(), w2(), w2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(wResult), precision,
            err_msg='Wrong Weibull random number sequence generated with duplicate')
    testResult = (w3(), w3(), w3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(wResult), precision,
            err_msg='Wrong Weibull random number sequence generated from serialize')

    # Check that the mean and variance come out right
    w = galsim.WeibullDeviate(testseed, a=wA, b=wB)
    vals = [w() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    gammaFactor1 = math.gamma(1.+1./wA)
    gammaFactor2 = math.gamma(1.+2./wA)
    mu = wB * gammaFactor1
    v = wB**2 * gammaFactor2 - mu**2
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from WeibullDeviate')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from WeibullDeviate')

    # Check discard
    w2 = galsim.WeibullDeviate(testseed, a=wA, b=wB)
    w2.discard(nvals)
    v1,v2 = w(),w2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    assert w.has_reliable_discard
    assert not w.generates_in_pairs

    # Check seed, reset
    w.seed(testseed)
    testResult2 = (w(), w(), w())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated after seed')

    w.reset(testseed)
    testResult2 = (w(), w(), w())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    w.reset(rng)
    testResult2 = (w(), w(), w())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    w.reset(ud)
    testResult = (w(), w(), w())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated after reset(ud)')

    # Check that two connected weibull deviates work correctly together.
    w2 = galsim.WeibullDeviate(testseed, a=wA, b=wB)
    w.reset(w2)
    testResult2 = (w(), w2(), w())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated using two wds')
    w.seed(testseed)
    testResult2 = (w2(), w(), w2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong weibull random number sequence generated using two wds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    w.seed()
    testResult2 = (w(), w(), w())
    assert testResult2 != testResult
    w.reset()
    testResult3 = (w(), w(), w())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    w.reset()
    testResult4 = (w(), w(), w())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    w = galsim.WeibullDeviate(a=wA, b=wB)
    testResult5 = (w(), w(), w())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Test generate
    w.seed(testseed)
    test_array = np.empty(3)
    w.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(wResult), precision,
            err_msg='Wrong weibull random number sequence from generate.')

    # Test generate with a float32 array
    w.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    w.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(wResult), precisionF,
            err_msg='Wrong weibull random number sequence from generate.')

    # Check that generated values are independent of number of threads.
    w1 = galsim.WeibullDeviate(testseed, a=3.1, b=7.3)
    w2 = galsim.WeibullDeviate(testseed, a=3.1, b=7.3)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        w1.generate(v1)
    with single_threaded(num_threads=10):
        w2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        w1.add_generate(v1)
    with single_threaded(num_threads=10):
        w2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(w, lambda x: (x.serialize(), x.a, x.b), random=True)
    check_pickle(w, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(w, random=True)
    assert 'WeibullDeviate' in repr(w)
    assert 'WeibullDeviate' in str(w)
    assert isinstance(eval(repr(w)), galsim.WeibullDeviate)
    assert isinstance(eval(str(w)), galsim.WeibullDeviate)

    # Check that we can construct a WeibullDeviate from None, and that it depends on dev/random.
    w1 = galsim.WeibullDeviate(None)
    w2 = galsim.WeibullDeviate(None)
    assert w1 != w2, "Consecutive WeibullDeviate(None) compared equal!"
    # We shouldn't be able to construct a WeibullDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.WeibullDeviate, dict())
    assert_raises(TypeError, galsim.WeibullDeviate, list())
    assert_raises(TypeError, galsim.WeibullDeviate, set())


@timer
def test_gamma():
    """Test Gamma random number generator
    """
    g = galsim.GammaDeviate(testseed, k=gammaK, theta=gammaTheta)
    g2 = g.duplicate()
    g3 = galsim.GammaDeviate(g.serialize(), k=gammaK, theta=gammaTheta)
    testResult = (g(), g(), g())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gammaResult), precision,
            err_msg='Wrong Gamma random number sequence generated')
    testResult = (g2(), g2(), g2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gammaResult), precision,
            err_msg='Wrong Gamma random number sequence generated with duplicate')
    testResult = (g3(), g3(), g3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(gammaResult), precision,
            err_msg='Wrong Gamma random number sequence generated from serialize')

    # Check that the mean and variance come out right
    g = galsim.GammaDeviate(testseed, k=gammaK, theta=gammaTheta)
    vals = [g() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = gammaK*gammaTheta
    v = gammaK*gammaTheta**2
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from GammaDeviate')
    np.testing.assert_almost_equal(var, v, 0,
            err_msg='Wrong variance from GammaDeviate')

    # Check discard
    g2 = galsim.GammaDeviate(testseed, k=gammaK, theta=gammaTheta)
    g2.discard(nvals, suppress_warnings=True)
    v1,v2 = g(),g2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    # Gamma uses at least 2 rngs per value, but can use arbitrarily more than this.
    assert v1 != v2
    assert not g.has_reliable_discard
    assert not g.generates_in_pairs

    # Discard normally emits a warning for Gamma
    g2 = galsim.GammaDeviate(testseed, k=gammaK, theta=gammaTheta)
    with assert_warns(galsim.GalSimWarning):
        g2.discard(nvals)

    # Check seed, reset
    g.seed(testseed)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated after seed')

    g.reset(testseed)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    g.reset(rng)
    testResult2 = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    g.reset(ud)
    testResult = (g(), g(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated after reset(ud)')

    # Check that two connected gamma deviates work correctly together.
    g2 = galsim.GammaDeviate(testseed, k=gammaK, theta=gammaTheta)
    g.reset(g2)
    testResult2 = (g(), g2(), g())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated using two gds')
    g.seed(testseed)
    testResult2 = (g2(), g(), g2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong gamma random number sequence generated using two gds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    g.seed()
    testResult2 = (g(), g(), g())
    assert testResult2 != testResult
    g.reset()
    testResult3 = (g(), g(), g())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    g.reset()
    testResult4 = (g(), g(), g())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    g = galsim.GammaDeviate(k=gammaK, theta=gammaTheta)
    testResult5 = (g(), g(), g())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Test generate
    g.seed(testseed)
    test_array = np.empty(3)
    g.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gammaResult), precision,
            err_msg='Wrong gamma random number sequence from generate.')

    # Test generate with a float32 array
    g.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    g.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(gammaResult), precisionF,
            err_msg='Wrong gamma random number sequence from generate.')

    # Check picklability
    check_pickle(g, lambda x: (x.serialize(), x.k, x.theta), random=True)
    check_pickle(g, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(g, random=True)
    assert 'GammaDeviate' in repr(g)
    assert 'GammaDeviate' in str(g)
    assert isinstance(eval(repr(g)), galsim.GammaDeviate)
    assert isinstance(eval(str(g)), galsim.GammaDeviate)

    # Check that we can construct a GammaDeviate from None, and that it depends on dev/random.
    g1 = galsim.GammaDeviate(None)
    g2 = galsim.GammaDeviate(None)
    assert g1 != g2, "Consecutive GammaDeviate(None) compared equal!"
    # We shouldn't be able to construct a GammaDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.GammaDeviate, dict())
    assert_raises(TypeError, galsim.GammaDeviate, list())
    assert_raises(TypeError, galsim.GammaDeviate, set())


@timer
def test_chi2():
    """Test Chi^2 random number generator
    """
    c = galsim.Chi2Deviate(testseed, n=chi2N)
    c2 = c.duplicate()
    c3 = galsim.Chi2Deviate(c.serialize(), n=chi2N)
    testResult = (c(), c(), c())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(chi2Result), precision,
            err_msg='Wrong Chi^2 random number sequence generated')
    testResult = (c2(), c2(), c2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(chi2Result), precision,
            err_msg='Wrong Chi^2 random number sequence generated with duplicate')
    testResult = (c3(), c3(), c3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(chi2Result), precision,
            err_msg='Wrong Chi^2 random number sequence generated from serialize')

    # Check that the mean and variance come out right
    c = galsim.Chi2Deviate(testseed, n=chi2N)
    vals = [c() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = chi2N
    v = 2.*chi2N
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from Chi2Deviate')
    np.testing.assert_almost_equal(var, v, 0,
            err_msg='Wrong variance from Chi2Deviate')

    # Check discard
    c2 = galsim.Chi2Deviate(testseed, n=chi2N)
    c2.discard(nvals, suppress_warnings=True)
    v1,v2 = c(),c2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    # Chi2 uses at least 2 rngs per value, but can use arbitrarily more than this.
    assert v1 != v2
    assert not c.has_reliable_discard
    assert not c.generates_in_pairs

    # Discard normally emits a warning for Chi2
    c2 = galsim.Chi2Deviate(testseed, n=chi2N)
    with assert_warns(galsim.GalSimWarning):
        c2.discard(nvals)

    # Check seed, reset
    c.seed(testseed)
    testResult2 = (c(), c(), c())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated after seed')

    c.reset(testseed)
    testResult2 = (c(), c(), c())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    c.reset(rng)
    testResult2 = (c(), c(), c())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    c.reset(ud)
    testResult = (c(), c(), c())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated after reset(ud)')

    # Check that two connected Chi^2 deviates work correctly together.
    c2 = galsim.Chi2Deviate(testseed, n=chi2N)
    c.reset(c2)
    testResult2 = (c(), c2(), c())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated using two cds')
    c.seed(testseed)
    testResult2 = (c2(), c(), c2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong Chi^2 random number sequence generated using two cds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    c.seed()
    testResult2 = (c(), c(), c())
    assert testResult2 != testResult
    c.reset()
    testResult3 = (c(), c(), c())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    c.reset()
    testResult4 = (c(), c(), c())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    c = galsim.Chi2Deviate(n=chi2N)
    testResult5 = (c(), c(), c())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Test generate
    c.seed(testseed)
    test_array = np.empty(3)
    c.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(chi2Result), precision,
            err_msg='Wrong Chi^2 random number sequence from generate.')

    # Test generate with a float32 array
    c.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    c.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(chi2Result), precisionF,
            err_msg='Wrong Chi^2 random number sequence from generate.')

    # Check picklability
    check_pickle(c, lambda x: (x.serialize(), x.n), random=True)
    check_pickle(c, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(c, random=True)
    assert 'Chi2Deviate' in repr(c)
    assert 'Chi2Deviate' in str(c)
    assert isinstance(eval(repr(c)), galsim.Chi2Deviate)
    assert isinstance(eval(str(c)), galsim.Chi2Deviate)

    # Check that we can construct a Chi2Deviate from None, and that it depends on dev/random.
    c1 = galsim.Chi2Deviate(None)
    c2 = galsim.Chi2Deviate(None)
    assert c1 != c2, "Consecutive Chi2Deviate(None) compared equal!"
    # We shouldn't be able to construct a Chi2Deviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.Chi2Deviate, dict())
    assert_raises(TypeError, galsim.Chi2Deviate, list())
    assert_raises(TypeError, galsim.Chi2Deviate, set())


@timer
def test_distfunction():
    """Test distribution-defined random number generator with a function
    """
    # Make sure it requires an input function in order to work.
    assert_raises(TypeError, galsim.DistDeviate)
    # Make sure it does appropriate input sanity checks.
    assert_raises(TypeError, galsim.DistDeviate,
                  function='../examples/data/cosmo-fid.zmed1.00_smoothed.out',
                  x_min=1.)
    assert_raises(TypeError, galsim.DistDeviate, function=1.0)
    assert_raises(ValueError, galsim.DistDeviate, function='foo.dat')
    assert_raises(TypeError, galsim.DistDeviate, function = lambda x : x*x, interpolant='linear')
    assert_raises(TypeError, galsim.DistDeviate, function = lambda x : x*x)
    assert_raises(TypeError, galsim.DistDeviate, function = lambda x : x*x, x_min=1.)
    test_vals = range(10)
    assert_raises(TypeError, galsim.DistDeviate,
                  function=galsim.LookupTable(test_vals, test_vals),
                  x_min = 1.)
    foo = galsim.DistDeviate(10, galsim.LookupTable(test_vals, test_vals))
    assert_raises(ValueError, foo.val, -1.)
    assert_raises(ValueError, galsim.DistDeviate, function = lambda x : -1, x_min=dmin, x_max=dmax)
    assert_raises(ValueError, galsim.DistDeviate, function = lambda x : x**2-1, x_min=dmin, x_max=dmax)

    d = galsim.DistDeviate(testseed, function=dfunction, x_min=dmin, x_max=dmax)
    d2 = d.duplicate()
    d3 = galsim.DistDeviate(d.serialize(), function=dfunction, x_min=dmin, x_max=dmax)
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated')
    testResult = (d2(), d2(), d2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated with duplicate')
    testResult = (d3(), d3(), d3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated from serialize')

    # Check val() method
    # pdf(x) = x^2
    # cdf(x) = (x/2)^3
    # val(y) = 2 y^(1/3)
    np.testing.assert_almost_equal(d.val(0), 0, 4)
    np.testing.assert_almost_equal(d.val(1), 2, 4)
    np.testing.assert_almost_equal(d.val(0.125), 1, 4)
    np.testing.assert_almost_equal(d.val(0.027), 0.6, 4)
    np.testing.assert_almost_equal(d.val(0.512), 1.6, 4)
    u = galsim.UniformDeviate(testseed)
    testResult = (d.val(u()), d.val(u()), d.val(u()))
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate sequence using d.val(u())')

    # Check that the mean and variance come out right
    d = galsim.DistDeviate(testseed, function=dfunction, x_min=dmin, x_max=dmax)
    vals = [d() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = 3./2.
    v = 3./20.
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from DistDeviate random numbers using function')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from DistDeviate random numbers using function')

    # Check discard
    d2 = galsim.DistDeviate(testseed, function=dfunction, x_min=dmin, x_max=dmax)
    d2.discard(nvals)
    v1,v2 = d(),d2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    assert d.has_reliable_discard
    assert not d.generates_in_pairs

    # Check seed, reset
    d.seed(testseed)
    testResult2 = (d(), d(), d())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated after seed')

    d.reset(testseed)
    testResult2 = (d(), d(), d())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated after reset(seed)')

    rng = galsim.BaseDeviate(testseed)
    d.reset(rng)
    testResult2 = (d(), d(), d())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated after reset(rng)')

    ud = galsim.UniformDeviate(testseed)
    d.reset(ud)
    testResult = (d(), d(), d())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated after reset(ud)')

    # Check that two connected DistDeviate deviates work correctly together.
    d2 = galsim.DistDeviate(testseed, function=dfunction, x_min=dmin, x_max=dmax)
    d.reset(d2)
    testResult2 = (d(), d2(), d())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated using two dds')
    d.seed(testseed)
    testResult2 = (d2(), d(), d2())
    np.testing.assert_array_equal(
            np.array(testResult), np.array(testResult2),
            err_msg='Wrong DistDeviate random number sequence generated using two dds after seed')

    # Check that seeding with the time works (although we cannot check the output).
    # We're mostly just checking that this doesn't raise an exception.
    # The output could be anything.
    d.seed()
    testResult2 = (d(), d(), d())
    assert testResult2 != testResult
    d.reset()
    testResult3 = (d(), d(), d())
    assert testResult3 != testResult
    assert testResult3 != testResult2
    d.reset()
    testResult4 = (d(), d(), d())
    assert testResult4 != testResult
    assert testResult4 != testResult2
    assert testResult4 != testResult3
    d = galsim.DistDeviate(function=dfunction, x_min=dmin, x_max=dmax)
    testResult5 = (d(), d(), d())
    assert testResult5 != testResult
    assert testResult5 != testResult2
    assert testResult5 != testResult3
    assert testResult5 != testResult4

    # Check with lambda function
    d = galsim.DistDeviate(testseed, function=lambda x: x*x, x_min=dmin, x_max=dmax)
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated with lambda function')

    # Check auto-generated lambda function
    d = galsim.DistDeviate(testseed, function='x*x', x_min=dmin, x_max=dmax)
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated with auto-lambda function')

    # Test generate
    d.seed(testseed)
    test_array = np.empty(3)
    d.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence from generate.')

    # Test add_generate
    d.seed(testseed)
    d.add_generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, 2*np.array(dFunctionResult), precision,
            err_msg='Wrong DistDeviate random number sequence from add_generate.')

    # Test generate with a float32 array
    d.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    d.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(dFunctionResult), precisionF,
            err_msg='Wrong DistDeviate random number sequence from generate.')

    # Test add_generate with a float32 array
    d.seed(testseed)
    d.add_generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, 2*np.array(dFunctionResult), precisionF,
            err_msg='Wrong DistDeviate random number sequence from add_generate.')

    # Check that generated values are independent of number of threads.
    d1 = galsim.DistDeviate(testseed, function=lambda x: np.exp(-x**3), x_min=0, x_max=2)
    d2 = galsim.DistDeviate(testseed, function=lambda x: np.exp(-x**3), x_min=0, x_max=2)
    v1 = np.empty(555)
    v2 = np.empty(555)
    with single_threaded():
        d1.generate(v1)
    with single_threaded(num_threads=10):
        d2.generate(v2)
    np.testing.assert_array_equal(v1, v2)
    with single_threaded():
        d1.add_generate(v1)
    with single_threaded(num_threads=10):
        d2.add_generate(v2)
    np.testing.assert_array_equal(v1, v2)

    # Check picklability
    check_pickle(d, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(d, random=True)
    assert 'DistDeviate' in repr(d)
    assert 'DistDeviate' in str(d)
    assert isinstance(eval(repr(d)), galsim.DistDeviate)
    assert isinstance(eval(str(d)), galsim.DistDeviate)

    # Check that we can construct a DistDeviate from None, and that it depends on dev/random.
    c1 = galsim.DistDeviate(None, lambda x:1, 0, 1)
    c2 = galsim.DistDeviate(None, lambda x:1, 0, 1)
    assert c1 != c2, "Consecutive DistDeviate(None) compared equal!"
    # We shouldn't be able to construct a DistDeviate from anything but a BaseDeviate, int, str,
    # or None.
    assert_raises(TypeError, galsim.DistDeviate, dict(), lambda x:1, 0, 1)
    assert_raises(TypeError, galsim.DistDeviate, list(), lambda x:1, 0, 1)
    assert_raises(TypeError, galsim.DistDeviate, set(), lambda x:1, 0, 1)


@timer
def test_distLookupTable():
    """Test distribution-defined random number generator with a LookupTable
    """
    precision = 9
    # Note: 256 used to be the default, so this is a regression test
    # We check below that it works with the default npoints=None
    d = galsim.DistDeviate(testseed, function=dLookupTable, npoints=256)
    d2 = d.duplicate()
    d3 = galsim.DistDeviate(d.serialize(), function=dLookupTable, npoints=256)
    np.testing.assert_equal(
            d.x_min, dLookupTable.x_min,
            err_msg='DistDeviate and the LookupTable passed to it have different lower bounds')
    np.testing.assert_equal(
            d.x_max, dLookupTable.x_max,
            err_msg='DistDeviate and the LookupTable passed to it have different upper bounds')

    testResult = (d(), d(), d())
    print('testResult = ',testResult)
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence using LookupTable')
    testResult = (d2(), d2(), d2())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence using LookupTable with duplicate')
    testResult = (d3(), d3(), d3())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence using LookupTable from serialize')

    # Check that the mean and variance come out right
    d = galsim.DistDeviate(testseed, function=dLookupTable, npoints=256)
    vals = [d() for i in range(nvals)]
    mean = np.mean(vals)
    var = np.var(vals)
    mu = 2.
    v = 7./3.
    print('mean = ',mean,'  true mean = ',mu)
    print('var = ',var,'   true var = ',v)
    np.testing.assert_almost_equal(mean, mu, 1,
            err_msg='Wrong mean from DistDeviate random numbers using LookupTable')
    np.testing.assert_almost_equal(var, v, 1,
            err_msg='Wrong variance from DistDeviate random numbers using LookupTable')

    # Check discard
    d2 = galsim.DistDeviate(testseed, function=dLookupTable, npoints=256)
    d2.discard(nvals)
    v1,v2 = d(),d2()
    print('after %d vals, next one is %s, %s'%(nvals,v1,v2))
    assert v1 == v2
    assert d.has_reliable_discard
    assert not d.generates_in_pairs

    # This should give the same values with only 5 points because of the particular nature
    # of these arrays.
    d = galsim.DistDeviate(testseed, function=dLookupTable, npoints=5)
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence for LookupTable with 5 points')

    # And it should also work if npoints is None
    d = galsim.DistDeviate(testseed, function=dLookupTable)
    testResult = (d(), d(), d())
    assert len(dLookupTable.x) == len(d._inverse_cdf.x)
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence for LookupTable with npoints=None')

    # Also read these values from a file
    d = galsim.DistDeviate(testseed, function=dLookupTableFile, interpolant='linear')
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence for LookupTable from file')

    d = galsim.DistDeviate(testseed, function=dLookupTableFile)
    testResult = (d(), d(), d())
    np.testing.assert_array_almost_equal(
            np.array(testResult), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence for LookupTable with default '
            'interpolant')

    # Test generate
    d.seed(testseed)
    test_array = np.empty(3)
    d.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence from generate.')

    # Test clip_neg
    # Same values as d, but with some negative values, where it had zeros.
    x = [0.0, 1.0, 1.000000001, 1.5, 2, 2.5, 2.999999999, 3.0, 4.0]
    f = [0.1, 0.1, 0.0, -0.2, -0.3, -0.2, 0.0    , 0.1, 0.1]
    tab2 = galsim.LookupTable(x=x, f=f, interpolant='linear')
    d2 = galsim.DistDeviate(testseed, function=tab2, clip_neg=True)
    test_array2 = np.empty(3)
    d2.generate(test_array2)
    np.testing.assert_array_almost_equal(test_array2, test_array)

    # Without clip_neg, it raises an error
    with assert_raises(galsim.GalSimValueError):
        galsim.DistDeviate(testseed, function=tab2)
    with assert_raises(galsim.GalSimValueError):
        galsim.DistDeviate(testseed, function=tab2, clip_neg=False)

    # Also works to cleanup NaN values
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        f = [0.1, 0.1, 0.0, -0.2, np.sqrt(-0.3), -0.2, 0.0    , 0.1, 0.1]
    tab3 = galsim.LookupTable(x=x, f=f, interpolant='linear')
    d3 = galsim.DistDeviate(testseed, function=tab3, clip_neg=True)
    test_array3 = np.empty(3)
    d3.generate(test_array3)
    np.testing.assert_array_almost_equal(test_array3, test_array)

    # Test filling an image
    d.seed(testseed)
    testimage = galsim.ImageD(np.zeros((3, 1)))
    testimage.addNoise(galsim.DeviateNoise(d))
    np.testing.assert_array_almost_equal(
            testimage.array.flatten(), np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence generated when applied to image.')

    # Test generate
    d.seed(testseed)
    test_array = np.empty(3)
    d.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(dLookupTableResult), precision,
            err_msg='Wrong DistDeviate random number sequence from generate.')

    # Test generate with a float32 array
    d.seed(testseed)
    test_array = np.empty(3, dtype=np.float32)
    d.generate(test_array)
    np.testing.assert_array_almost_equal(
            test_array, np.array(dLookupTableResult), precisionF,
            err_msg='Wrong DistDeviate random number sequence from generate.')

    # Test a case with nearly flat probabilities
    # x and p arrays with and without a small (epsilon) step
    dx_eps = np.arange(6)
    dp1_eps = np.zeros(dx_eps.shape)
    dp2_eps = np.zeros(dx_eps.shape)
    eps = np.finfo(dp1_eps[0].dtype).eps
    dp1_eps[0] = 0.5
    dp2_eps[0] = 0.5
    dp1_eps[-1] = 0.5
    dp2_eps[-2] = eps
    dp2_eps[-1] = 0.5-eps
    dLookupTableEps1 = galsim.LookupTable(x=dx_eps, f=dp1_eps, interpolant='linear')
    dLookupTableEps2 = galsim.LookupTable(x=dx_eps, f=dp2_eps, interpolant='linear')
    d1 = galsim.DistDeviate(testseed, function=dLookupTableEps1, npoints=len(dx_eps))
    d2 = galsim.DistDeviate(testseed, function=dLookupTableEps2, npoints=len(dx_eps))
    # If these were successfully created everything is probably fine, but check they create the same
    # internal LookupTable
    np.testing.assert_array_almost_equal(
            d1._inverse_cdf.getArgs(), d2._inverse_cdf.getArgs(), precision,
            err_msg='DistDeviate with near-flat probabilities incorrectly created '
                    'a monotonic version of the CDF')
    np.testing.assert_array_almost_equal(
            d1._inverse_cdf.getVals(), d2._inverse_cdf.getVals(), precision,
            err_msg='DistDeviate with near-flat probabilities incorrectly created '
                    'a monotonic version of the CDF')

    # And that they generate the same values
    ar1 = np.empty(100); d1.generate(ar1)
    ar2 = np.empty(100); d2.generate(ar2)
    np.testing.assert_array_almost_equal(ar1, ar2, precision,
            err_msg='Two DistDeviates with near-flat probabilities generated different values.')

    # Check picklability
    check_pickle(d, lambda x: (x(), x(), x(), x()), random=True)
    check_pickle(d, random=True)
    assert 'DistDeviate' in repr(d)
    assert 'DistDeviate' in str(d)
    assert isinstance(eval(repr(d)), galsim.DistDeviate)
    assert isinstance(eval(str(d)), galsim.DistDeviate)


@timer
def test_multiprocess():
    """Test that the same random numbers are generated in single-process and multi-process modes.
    """
    from multiprocessing import current_process
    from multiprocessing import get_context
    ctx = get_context('fork')
    Process = ctx.Process
    Queue = ctx.Queue

    def generate_list(seed):
        """Given a particular seed value, generate a list of random numbers.
           Should be deterministic given the input seed value.
        """
        rng = galsim.UniformDeviate(seed)
        out = []
        for i in range(20):
            out.append(rng())
        return out

    def worker(input, output):
        """input is a queue with seed values
           output is a queue storing the results of the tasks along with the process name,
           and which args the result is for.
        """
        for args in iter(input.get, 'STOP'):
            result = generate_list(*args)
            output.put( (result, current_process().name, args) )

    # Use sequential numbers.
    # On inspection, can see that even the first value in each list is random with
    # respect to the other lists.  i.e. "nearby" inputs do not produce nearby outputs.
    # I don't know of an actual assert to do for this, but it is clearly true.
    seeds = [ 1532424 + i for i in range(16) ]

    nproc = 4  # Each process will do 4 lists (typically)

    with single_threaded():
        # First make lists in the single process:
        ref_lists = dict()
        for seed in seeds:
            list = generate_list(seed)
            ref_lists[seed] = list

        # Now do this with multiprocessing
        # Put the seeds in a queue
        task_queue = Queue()
        for seed in seeds:
            task_queue.put( [seed] )

        # Run the tasks:
        done_queue = Queue()
        for k in range(nproc):
            Process(target=worker, args=(task_queue, done_queue)).start()

        # Check the results in the order they finished
        for i in range(len(seeds)):
            list, proc, args = done_queue.get()
            seed = args[0]
            np.testing.assert_array_equal(
                    list, ref_lists[seed],
                    err_msg="Random numbers are different when using multiprocessing")

        # Stop the processes:
        for k in range(nproc):
            task_queue.put('STOP')


@timer
def test_permute():
    """Simple tests of the permute() function."""
    # Make a fake list, and another list consisting of indices.
    my_list = [3.7, 4.1, 1.9, 11.1, 378.3, 100.0]
    import copy
    my_list_copy = copy.deepcopy(my_list)
    n_list = len(my_list)
    ind_list = list(range(n_list))

    # Permute both at the same time.
    galsim.random.permute(312, my_list, ind_list)

    # Make sure that everything is sensible
    for ind in range(n_list):
        assert my_list_copy[ind_list[ind]] == my_list[ind]

    # Repeat with same seed, should do same permutation.
    my_list = copy.deepcopy(my_list_copy)
    galsim.random.permute(312, my_list)
    for ind in range(n_list):
        assert my_list_copy[ind_list[ind]] == my_list[ind]

    # permute with no lists should raise TypeError
    with assert_raises(TypeError):
        galsim.random.permute(312)


@timer
def test_ne():
    """ Check that inequality works as expected for corner cases where the reprs of two
    unequal BaseDeviates may be the same due to truncation.
    """
    a = galsim.BaseDeviate(seed='1 2 3 4 5 6 7 8 9 10')
    b = galsim.BaseDeviate(seed='1 2 3 7 6 5 4 8 9 10')
    assert repr(a) == repr(b)
    assert a != b

    # Check DistDeviate separately, since it overrides __repr__ and __eq__
    d1 = galsim.DistDeviate(seed=a, function=galsim.LookupTable([1, 2, 3], [4, 5, 6]))
    d2 = galsim.DistDeviate(seed=b, function=galsim.LookupTable([1, 2, 3], [4, 5, 6]))
    assert repr(d1) == repr(d2)
    assert d1 != d2

@timer
def test_int64():
    # cf. #1009
    # Check that various possible integer types work as seeds.

    rng1 = galsim.BaseDeviate(int(123))
    # cf. https://www.numpy.org/devdocs/user/basics.types.html
    ivalues =[np.int8(123),  # Note this one requires i < 128
              np.int16(123),
              np.int32(123),
              np.int64(123),
              np.uint8(123),
              np.uint16(123),
              np.uint32(123),
              np.uint64(123),
              np.short(123),
              np.ushort(123),
              np.intc(123),
              np.uintc(123),
              np.intp(123),
              np.uintp(123),
              np.int_(123),
              np.longlong(123),
              np.ulonglong(123),
              np.array(123).astype(np.int64)]

    for i in ivalues:
        rng2 = galsim.BaseDeviate(i)
        assert rng2 == rng1

@timer
def test_numpy_generator():
    rng = galsim.BaseDeviate(1234)
    gen = galsim.BaseDeviate(1234).as_numpy_generator()

    # The regular (and somewhat cumbersome) GalSim way:
    a1 = np.empty(10, dtype=float)
    galsim.UniformDeviate(rng).generate(a1)
    a1 *= 9.
    a1 += 1.

    # The nicer numpy syntax
    a2 = gen.uniform(1.,10., size=10)
    print('a1 = ',a1)
    print('a2 = ',a2)
    np.testing.assert_array_equal(a1, a2)

    # Can also use the np property as a quick shorthand
    a1 = rng.np.normal(0, 10, size=20)
    a2 = gen.normal(0, 10, size=20)
    print('a1 = ',a1)
    print('a2 = ',a2)
    np.testing.assert_array_equal(a1, a2)

    # Check that normal gives statistically the right mean/var.
    # (Numpy's normal uses the next_uint64 function, so this is a non-trivial test of that
    # code, which I originally got wrong.)
    a3 = gen.normal(17, 23, size=1_000_000)
    print('mean = ',np.mean(a3))
    print('std = ',np.std(a3))
    assert np.isclose(np.mean(a3), 17, rtol=1.e-3)
    assert np.isclose(np.std(a3), 23, rtol=3.e-3)

if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import math

from .stamp import StampBuilder, RegisterStampType
from .value import ParseValue, GetAllParams, SetDefaultIndex
from .gsobject import BuildGSObject, TransformObject
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..convolve import Convolve
from ..angle import Angle, radians

# This file adds stamp type Ring which builds an object once every n times, and then
# rotates it in a ring for the other n-1 times per per group.

class RingBuilder(StampBuilder):
    """This performs the tasks necessary for building a Ring stamp type.

    It uses the regular Basic functions for most things.
    It specializes the setup, buildProfile, reject, and makeTasks functions.
    """

    def setup(self, config, base, xsize, ysize, ignore, logger):
        """Do the initialization and setup for the Ring type.

        Most of the work is done by SetupBasic, but we do need to check that the required parameters
        are present, and also that no additional parameters are present.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            xsize:      The xsize of the image to build (if known).
            ysize:      The ysize of the image to build (if known).
            ignore:     A list of parameters that are allowed to be in config that we can
                        ignore here. i.e. it won't be an error if these parameters are present.
            logger:     If given, a logger object to log progress.

        Returns:
            xsize, ysize, image_pos, world_pos
        """
        req = { 'num' : int }
        opt = { 'full_rotation' : Angle , 'index' : int }
        # Ignore the transformation specifications that are allowed in stamp for Ring types.
        ignore = ignore + [
            'dilate', 'dilation', 'ellip', 'rotate', 'rotation', 'scale_flux',
            'magnify', 'magnification', 'shear', 'shift' ]

        params = GetAllParams(config, base, req=req, opt=opt, ignore=ignore)[0]

        num = params['num']
        if num <= 0:
            raise GalSimConfigValueError("Attribute num for gal.type == Ring must be > 0", num)

        # Setup the indexing sequence if it hasn't been specified using the number of items.
        SetDefaultIndex(config, num)

        # Set the default full_rotation to pi radians
        if 'full_rotation' not in params:
            config['full_rotation'] = math.pi * radians

        # Now go on and do the base class setup.
        ignore = ignore + list(req) + list(opt)
        return super(RingBuilder, self).setup(config, base, xsize, ysize, ignore, logger)

    def buildProfile(self, config, base, psf, gsparams, logger):
        """
        Build the object to be drawn.

        For the first item in the ring, this is the same as Basic. It stores the galaxy object
        created on the first time.  Then for later stamps in the ring, it retrieves the stored
        first galaxy and just rotates it before convolving by the psf.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            psf:        The PSF, if any.  This may be None, in which case, no PSF is convolved.
            gsparams:   A dict of kwargs to use for a GSParams.  More may be added to this
                        list by the galaxy object.
            logger:     If given, a logger object to log progress.

        Returns:
            the final profile
        """
        # These have all already been checked to exist in SetupRing.
        num = ParseValue(config, 'num', base, int)[0]
        index = ParseValue(config, 'index', base, int)[0]
        if index < 0 or index >= num:
            raise GalSimConfigError("index %d out of bounds for Ring"%index)

        if index % num == 0:
            # Then we are on the first item in the ring, so make it normally.
            gal = BuildGSObject(base, 'gal', gsparams=gsparams, logger=logger)[0]
            if gal is None:
                raise GalSimConfigError(
                    "The gal field must define a valid galaxy for stamp type=Ring.")
            # Save the galaxy profile for next time.
            self.first = gal
        else:
            # Grab the saved first galaxy.
            if not hasattr(self, 'first'):
                raise GalSimConfigError(
                    "Building Ring after the first item, but no first gal stored.")
            gal = self.first
            full_rot = ParseValue(config, 'full_rotation', base, Angle)[0]
            dtheta = full_rot / num
            gal = gal.rotate(index*dtheta)

        # Apply any transformations that are given in the stamp field.
        gal = TransformObject(gal, config, base, logger)[0]

        if psf is not None:
            return Convolve(gal,psf)
        else:
            return gal

    def reject(self, config, base, prof, psf, image, logger):
        """Check to see if this object should be rejected.

        This is the same as base class reject for the first item in the ring.  Later items are not
        checked though, since rejecting them would mess up the ring.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            prof:       The profile that was drawn.
            psf:        The psf that was used to build the profile.
            image:      The postage stamp image.  No noise is on it yet at this point.
            logger:     If given, a logger object to log progress.

        Returns:
            whether the galaxy was rejected.
        """
        index = ParseValue(config, 'index', base, int)[0]
        if index == 0:
            return super(RingBuilder,self).reject(config, base, prof, psf, image, logger)
        else:
            return False

    def makeTasks(self, config, base, jobs, logger):
        """Turn a list of jobs into a list of tasks.

        For the Ring stamp type, we group jobs into sets of num.  If there are extra jobs that
        don't fit into a full ring, the last task will be a partial ring.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            jobs:       A list of jobs to split up into tasks.  Each job in the list is a
                        dict of parameters that includes 'obj_num'.
            logger:     If given, a logger object to log progress.

        Returns:
            a list of tasks
        """
        if 'num' not in config:
            raise GalSimConfigError("Attribute num is required for type = Ring")
        num = ParseValue(config, 'num', base, int)[0]
        ntot = len(jobs)
        tasks = [ [ (jobs[j], j) for j in range(k,min(k+num,ntot)) ] for k in range(0, ntot, num) ]
        return tasks


# Register this as a valid stamp type
RegisterStampType('Ring', RingBuilder())




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
import inspect

from .util import LoggerWrapper, GetIndex, GetRNG, get_cls_params, CleanConfig
from .value import ParseValue, GetCurrentValue, GetAllParams, CheckAllParams, SetDefaultIndex
from .input import RegisterInputConnectedType
from .sed import BuildSED
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..position import PositionD
from ..sum import Add
from ..convolve import Convolve
from ..phase_psf import OpticalPSF
from ..shear import Shear
from ..angle import Angle
from ..gsobject import GSObject
from ..chromatic import ChromaticObject, ChromaticOpticalPSF
from ..gsparams import GSParams
from ..utilities import basestring
from ..chromatic import ChromaticAtmosphere
from ..celestial import CelestialCoord

# This file handles the building of GSObjects in the config['psf'] and config['gal'] fields.
# This file includes many of the simple object types.  Additional types are defined in
# gsobject_ring.py, input_real.py, and input_cosmos.py.

# This module-level dict will store all the registered gsobject types.
# See the RegisterObjectType function at the end of this file.
# The keys will be the (string) names of the object types, and the values are the function
# to call to build an object of that type.
valid_gsobject_types = {}

class SkipThisObject(Exception):
    """
    A class that a builder can throw to indicate that nothing went wrong, but for some
    reason, this particular object should be skipped and just move onto the next object.
    The constructor takes an optional message that will be output to the logger if
    logging is active.
    """
    def __init__(self, message=None):
        # Using self.message gives a deprecation warning.  Avoid this by using a different name.
        self.msg = message


def BuildGSObject(config, key, base=None, gsparams={}, logger=None):
    """Build a GSObject from the parameters in config[key].

    Parameters:
        config:     A dict with the configuration information.
        key:        The key name in config indicating which object to build.
        base:       The base dict of the configuration. [default: config]
        gsparams:   Optionally, provide non-default GSParams items.  Any ``gsparams`` specified
                    at this level will be added to the list.  This should be a dict with
                    whatever kwargs should be used in constructing the GSParams object.
                    [default: {}]
        logger:     Optionally, provide a logger for logging debug statements.
                    [default: None]

    Returns:
        the tuple (gsobject, safe), where ``gsobject`` is the built object, and ``safe`` is
        a bool that says whether it is safe to use this object again next time.
    """
    from .. import __dict__ as galsim_dict

    logger = LoggerWrapper(logger)
    if base is None:
        base = config

    logger.debug('obj %d: Start BuildGSObject %s',base.get('obj_num',0),key)

    # If key isn't in config, then just return None.
    try:
        param = config[key]
    except KeyError:
        return None, True

    # Check what index key we want to use for this object.
    # Note: this call will also set base['index_key'] and base['rng'] to the right values
    index, index_key = GetIndex(param, base)

    # Get the type to be parsed.
    if not 'type' in param:
        raise GalSimConfigError("type attribute required in config.%s"%key)
    type_name = param['type']

    # If we are repeating, then we get to use the current object for repeat times.
    if 'repeat' in param:
        repeat = ParseValue(param, 'repeat', base, int)[0]
    else:
        repeat = 1

    # Check if we need to skip this object
    if 'skip' in param:
        skip = ParseValue(param, 'skip', base, bool)[0]
        if skip:
            logger.debug('obj %d: Skipping because field skip=True',base.get('obj_num',0))
            raise SkipThisObject()

    # Check if we can use the current cached object
    if 'current' in param:
        # NB. "current" tuple is (obj, safe, None, index, index_type)
        cobj, csafe, cvalue_type, cindex, cindex_type = param['current']
        if csafe or cindex//repeat == index//repeat:
            # If logging, explain why we are using the current object.
            if logger:
                if csafe:
                    logger.debug('obj %d: current is safe',base.get('obj_num',0))
                elif repeat > 1:
                    logger.debug('obj %d: repeat = %d, index = %d, use current object',
                                base.get('obj_num',0),repeat,index)
                else:
                    logger.debug('obj %d: This object is already current', base.get('obj_num',0))

            return cobj, csafe

    # Set up the initial default list of attributes to ignore while building the object:
    ignore = [
        'dilate', 'dilation', 'ellip', 'rotate', 'rotation', 'scale_flux',
        'magnify', 'magnification', 'shear', 'lens', 'shift', 'sed',
        'gsparams', 'skip',
        'current', 'index_key', 'repeat'
    ]
    # There are a few more that are specific to which key we have.
    # Note: some custom stamp builders may have fields besides just gal and psf.
    # Using 'gal' in key rather than key == 'gal', we make it easier for them, since the
    # keys can be e.g. blue_gal, red_gal, or halo_gal, field_gal, etc.  Anything with gal
    # somewhere in the name will be treated as a gal.  Likewise ground_psf, space_psf or
    # similar will all be treated as psf.
    if isinstance(key, basestring) and 'gal' in key:
        ignore += [ 'resolution', 'signal_to_noise', 'redshift', 're_from_res' ]
    elif isinstance(key, basestring) and 'psf' in key:
        ignore += [ 'saved_re' ]
    else:
        # As long as key isn't psf, allow resolution.
        # Ideally, we'd like to check that it's something within the gal hierarchy, but
        # I don't know an easy way to do that.
        ignore += [ 'resolution' , 're_from_res' ]

    # Allow signal_to_noise for PSFs only if there is not also a galaxy.
    if 'gal' not in base and isinstance(key, basestring) and 'psf' in key:
        ignore += [ 'signal_to_noise']

    # If we are specifying the size according to a resolution, then we
    # need to get the PSF's half_light_radius.
    if 'resolution' in param:
        if 'psf' not in base:
            raise GalSimConfigError("Cannot use gal.resolution if no psf is set.")
        if 'saved_re' not in base['psf']:
            raise GalSimConfigError(
                'Cannot use gal.resolution with psf.type = %s'%base['psf']['type'])
        psf_re = base['psf']['saved_re']
        resolution = ParseValue(param, 'resolution', base, float)[0]
        gal_re = resolution * psf_re
        if 're_from_res' not in param:
            # The first time, check that half_light_radius isn't also specified.
            if 'half_light_radius' in param:
                raise GalSimConfigError(
                    'Cannot specify both gal.resolution and gal.half_light_radius')
            param['re_from_res'] = True
        param['half_light_radius'] = gal_re

    if 'gsparams' in param:
        gsparams = UpdateGSParams(gsparams, param['gsparams'], base)

    # See if this type is registered as a valid type.
    if type_name in valid_gsobject_types:
        build_func = valid_gsobject_types[type_name]
    elif type_name in galsim_dict:
        gdict = globals().copy()
        exec('import galsim', gdict)
        build_func = eval("galsim."+type_name, gdict)
    else:
        raise GalSimConfigValueError("Unrecognised gsobject type", type_name)

    if inspect.isclass(build_func) and issubclass(build_func, (GSObject, ChromaticObject)):
        gsobject, safe = _BuildSimple(build_func, param, base, ignore, gsparams, logger)
    else:
        gsobject, safe = build_func(param, base, ignore, gsparams, logger)

    # Apply any SED and redshift that might be present.
    if 'redshift' in param:
        if 'sed' in param:
            from ..deprecated import depr
            depr('gal.redshift', '2.5.3', 'gal.sed.redshift',
                 'For chromatic objects, the redshift parameter should be given in the '
                 'sed field.')
            param['sed']['redshift'] = param.pop('redshift')

    gsobject, safe1 = ApplySED(gsobject, param, base, logger)
    safe = safe and safe1

    if 'redshift' in param:
        redshift, safe1 = ParseValue(param, 'redshift', base, float)
        safe = safe and safe1
        if isinstance(gsobject, ChromaticObject):
            from ..deprecated import depr
            depr('gal.redshift', '2.5.3', 'gal.sed.redshift',
                 'For chromatic objects, the redshift parameter should be given in the '
                 'sed field.')
            gsobject = gsobject._atRedshift(redshift)
        else:
            # If not chromatic, then redshift is just an attribute for reference.
            gsobject.redshift = redshift

    if 'flux' in param:
        flux, safe1 = ParseValue(param, 'flux', base, float)
        logger.debug('obj %d: flux == %f',base.get('obj_num',0),flux)
        if 'sed' in param and 'bandpass' in base:
            gsobject = gsobject.withFlux(flux, bandpass=base['bandpass'])
        else:
            gsobject = gsobject.withFlux(flux)
        safe = safe and safe1

    # If this is a psf, try to save the half_light_radius in case gal uses resolution.
    if key == 'psf':
        try:
            param['saved_re'] = gsobject.half_light_radius
        except (AttributeError, NotImplementedError, TypeError):
            pass

    # Apply any dilation, ellip, shear, etc. modifications.
    gsobject, safe1 = TransformObject(gsobject, param, base, logger)
    safe = safe and safe1

    # Re-get index and index_key in case something changed when building the object.
    # (cf. Roman PSF for an example of why this might be useful.)
    index, index_key = GetIndex(param, base)
    param['current'] = gsobject, safe, None, index, index_key

    return gsobject, safe


def UpdateGSParams(gsparams, config, base):
    """Add additional items to the ``gsparams`` dict based on config['gsparams'].

    Parameters:
        gsparams:   A dict with whatever kwargs should be used in constructing the GSParams object.
        config:     A dict with the configuration information.
        base:       The base dict of the configuration.

    Returns:
        an updated gsparams dict
    """
    opt = GSObject._gsparams_opt
    kwargs, safe = GetAllParams(config, base, opt=opt)
    # When we update gsparams, we don't want to corrupt the original, so we need to
    # make a copy first, then update with kwargs.
    ret = {}
    ret.update(gsparams)
    ret.update(kwargs)
    return ret

def ApplySED(gsobject, config, base, logger):
    """Read and apply an SED to the base gsobject

    Parameters:
        gsobject:   The base GSObject
        config:     A dict with the configuration information.
        base:       The base dict of the configuration.
        logger:     A logger for logging debug statements.
    """
    if 'sed' in config:
        sed, safe = BuildSED(config, 'sed', base, logger)
        return gsobject * sed, safe
    else:
        return gsobject, True

#
# The following are private functions to implement the simpler GSObject types.
# These are not imported into galsim.config namespace.
#

def _BuildSimple(build_func, config, base, ignore, gsparams, logger):
    """Build a simple GSObject (i.e. one without a specialized _Build function) or
    any other GalSim object that defines _req_params, _opt_params and _single_params.
    """
    # Build the kwargs according to the various params objects in the class definition.
    type_name = config['type']
    logger.debug('obj %d: BuildSimple for type = %s',base.get('obj_num',0),type_name)

    req, opt, single, takes_rng = get_cls_params(build_func)
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt, single=single, ignore=ignore)
    if gsparams: kwargs['gsparams'] = GSParams(**gsparams)

    if takes_rng:
        kwargs['rng'] = GetRNG(config, base, logger, type_name)
        safe = False

    logger.debug('obj %d: kwargs = %s',base.get('obj_num',0),kwargs)

    # Finally, after pulling together all the params, try making the GSObject.
    return build_func(**kwargs), safe


def _BuildNone(config, base, ignore, gsparams, logger):
    """Special type=None returns None.
    """
    return None, True


def _BuildAdd(config, base, ignore, gsparams, logger):
    """Build a Sum object.
    """
    req = { 'items' : list }
    opt = { 'flux' : float }
    # Only Check, not Get.  We need to handle items a bit differently, since it's a list.
    CheckAllParams(config, req=req, opt=opt, ignore=ignore)

    gsobjects = []
    items = config['items']
    if not isinstance(items,list):
        raise GalSimConfigError("items entry for type=Add is not a list.")
    safe = True

    for i in range(len(items)):
        gsobject, safe1 = BuildGSObject(items, i, base, gsparams, logger)
        # Skip items with flux=0
        if 'flux' in items[i] and GetCurrentValue('flux',items[i],float,base) == 0.:
            logger.debug('obj %d: Not including component with flux == 0',base.get('obj_num',0))
            continue
        safe = safe and safe1
        gsobjects.append(gsobject)

    if len(gsobjects) == 0:
        raise GalSimConfigError("No valid items for type=Add")
    elif len(gsobjects) == 1:
        gsobject = gsobjects[0]
    else:
        # Special: if the last item in a Sum doesn't specify a flux, we scale it
        # to bring the total flux up to 1.
        if ('flux' not in items[-1]) and all('flux' in item for item in items[0:-1]):
            sum_flux = 0
            for item in items[0:-1]:
                sum_flux += GetCurrentValue('flux',item,float,base)
            f = 1. - sum_flux
            if (f < 0):
                logger.warning(
                    "Warning: Automatic flux for the last item in Sum (to make the total flux=1) "
                    "resulted in negative flux = %f for that item"%f)
            logger.debug('obj %d: Rescaling final object in sum to have flux = %f',
                         base.get('obj_num',0), f)
            gsobjects[-1] = gsobjects[-1].withFlux(f)
        if gsparams: gsparams = GSParams(**gsparams)
        else: gsparams = None
        gsobject = Add(gsobjects,gsparams=gsparams)

    return gsobject, safe

def _BuildConvolve(config, base, ignore, gsparams, logger):
    """Build a Convolution object.
    """
    req = { 'items' : list }
    opt = { 'flux' : float }
    # Only Check, not Get.  We need to handle items a bit differently, since it's a list.
    CheckAllParams(config, req=req, opt=opt, ignore=ignore)

    gsobjects = []
    items = config['items']
    if not isinstance(items,list):
        raise GalSimConfigError("items entry for type=Convolve is not a list.")
    safe = True
    for i in range(len(items)):
        gsobject, safe1 = BuildGSObject(items, i, base, gsparams, logger)
        safe = safe and safe1
        gsobjects.append(gsobject)

    if len(gsobjects) == 0:
        raise GalSimConfigError("No valid items for type=Convolve")
    elif len(gsobjects) == 1:
        gsobject = gsobjects[0]
    else:
        if gsparams: gsparams = GSParams(**gsparams)
        else: gsparams = None
        gsobject = Convolve(gsobjects,gsparams=gsparams)

    return gsobject, safe

def _BuildList(config, base, ignore, gsparams, logger):
    """Build a GSObject selected from a List.
    """
    req = { 'items' : list }
    opt = { 'index' : float , 'flux' : float }
    # Only Check, not Get.  We need to handle items a bit differently, since it's a list.
    CheckAllParams(config, req=req, opt=opt, ignore=ignore)

    items = config['items']
    if not isinstance(items,list):
        raise GalSimConfigError("items entry for type=List is not a list.")

    # Setup the indexing sequence if it hasn't been specified using the length of items.
    SetDefaultIndex(config, len(items))
    index, safe = ParseValue(config, 'index', base, int)
    if index < 0 or index >= len(items):
        raise GalSimConfigError("index %d out of bounds for List"%index)

    gsobject, safe1 = BuildGSObject(items, index, base, gsparams, logger)
    safe = safe and safe1

    return gsobject, safe

def _BuildEval(config, base, ignore, gsparams, logger):
    """Build a GSObject from an Eval string
    """
    from .value_eval import _GenerateFromEval
    req = { 'str': str }
    params, _ = GetAllParams(config, base, req=req, ignore=ignore)
    gsobject, safe = _GenerateFromEval(params, base, None)
    if gsparams:
        gsobject = gsobject.withGSParams(**gsparams)

    return gsobject, safe


def ParseAberrations(key, config, base, name):
    """Parse a possible aberrations list in config dict.

    Parameters:
        key:        The key name with the aberrations list.
        config:     A dict with the tranformation information for this object.
        base:       The base dict of the configuration.
        name:       The name of the source object being parsed (only used for error reporting).

    Returns:
        aberrations list or None
    """
    if key in config:
        aber_list = [0.0] * 4  # Initial 4 values are ignored.
        aberrations = config[key]
        if not isinstance(aberrations,list):
            raise GalSimConfigError(
                "aberrations entry for config.%s entry is not a list."%(name))
        safe = True
        for i in range(len(aberrations)):
            value, safe1 = ParseValue(aberrations, i, base, float)
            aber_list.append(value)
            safe = safe and safe1
        return aber_list
    else:
        return None

def _BuildJointOpticalPSF(cls, config, base, ignore, gsparams, logger):
    req, opt, single, _ = get_cls_params(cls)

    kwargs, safe = GetAllParams(config, base, req, opt, single, ignore = ['aberrations'] + ignore)
    if gsparams: kwargs['gsparams'] = GSParams(**gsparams)
    kwargs['aberrations'] = ParseAberrations('aberrations', config, base, cls.__name__)

    return cls(**kwargs), safe

def _BuildOpticalPSF(config, base, ignore, gsparams, logger):
    """Build an OpticalPSF.
    """
    return _BuildJointOpticalPSF(OpticalPSF, config, base, ignore, gsparams, logger)

def _BuildChromaticOpticalPSF(config, base, ignore, gsparams, logger):
    """Build a ChromaticOpticalPSF.
    """
    # All the code for this is the same as for OpticalPSF, so use a shared implementation above.
    return _BuildJointOpticalPSF(ChromaticOpticalPSF, config, base, ignore, gsparams, logger)

def _BuildChromaticAtmosphere(config, base, ignore, gsparams, logger):
    """Build a ChromaticAtmosphere.
    """
    req = {'base_wavelength' : float}
    opt = {
           'alpha' : float,
           'zenith_angle' : Angle,
           'parallactic_angle' : Angle,
           'zenith_coord' : CelestialCoord,
           'HA' : Angle,
           'latitude' : Angle,
           'pressure' : float,
           'temperature' : float,
           'H2O_pressure' : float,
          }
    ignore = ['base_profile'] + ignore
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt, ignore=ignore)

    if 'base_profile' not in config:
        raise GalSimConfigError("Attribute base_profile is required for type=ChromaticAtmosphere")
    base_profile, safe1 = BuildGSObject(config, 'base_profile', base, gsparams, logger)
    safe = safe and safe1

    if 'zenith_angle' not in kwargs:
        sky_pos = base.get('sky_pos', None)
        if sky_pos is None:
            raise GalSimConfigError("Using zenith_angle with type=ChromaticAtmosphere requires "
                                    "that sky_pos be available to use as the object coord.")
        kwargs['obj_coord'] = sky_pos
        safe = False

    psf = ChromaticAtmosphere(base_profile, **kwargs)
    return psf, safe


#
# Now the functions for performing transformations
#

def TransformObject(gsobject, config, base, logger):
    """Applies ellipticity, rotation, gravitational shearing and centroid shifting to a
    supplied GSObject, in that order.

    Parameters:
        gsobject:   The GSObject to be transformed.
        config:     A dict with the tranformation information for this object.
        base:       The base dict of the configuration.
        logger:     A logger for logging debug statements.

    Returns:
        transformed GSObject.
    """
    logger = LoggerWrapper(logger)
    # The transformations are applied in the following order:
    _transformation_list = [
        ('dilate', _Dilate),
        ('dilation', _Dilate),
        ('ellip', _Shear),
        ('rotate', _Rotate),
        ('rotation', _Rotate),
        ('scale_flux', _ScaleFlux),
        ('lens', _Lens),
        ('shear', _Shear),
        ('magnify', _Magnify),
        ('magnification', _Magnify),
        ('shift', _Shift),
    ]

    safe = True
    for key, func in _transformation_list:
        if key in config:
            gsobject, safe1 = func(gsobject, config, key, base, logger)
            safe = safe and safe1
    return gsobject, safe

def _Shear(gsobject, config, key, base, logger):
    shear, safe = ParseValue(config, key, base, Shear)
    logger.debug('obj %d: shear = %f,%f',base.get('obj_num',0),shear.g1,shear.g2)
    gsobject = gsobject._shear(shear)
    return gsobject, safe

def _Rotate(gsobject, config, key, base, logger):
    theta, safe = ParseValue(config, key, base, Angle)
    logger.debug('obj %d: theta = %f rad',base.get('obj_num',0),theta.rad)
    gsobject = gsobject.rotate(theta)
    return gsobject, safe

def _ScaleFlux(gsobject, config, key, base, logger):
    flux_ratio, safe = ParseValue(config, key, base, float)
    logger.debug('obj %d: flux_ratio  = %f',base.get('obj_num',0),flux_ratio)
    gsobject = gsobject * flux_ratio
    return gsobject, safe

def _Dilate(gsobject, config, key, base, logger):
    scale, safe = ParseValue(config, key, base, float)
    logger.debug('obj %d: scale  = %f',base.get('obj_num',0),scale)
    gsobject = gsobject.dilate(scale)
    return gsobject, safe

def _Lens(gsobject, config, key, base, logger):
    shear, safe = ParseValue(config[key], 'shear', base, Shear)
    mu, safe1 = ParseValue(config[key], 'mu', base, float)
    safe = safe and safe1
    logger.debug('obj %d: shear = %f,%f',base.get('obj_num',0),shear.g1,shear.g2)
    logger.debug('obj %d: mu  = %f',base.get('obj_num',0),mu)
    gsobject = gsobject._lens(shear.g1, shear.g2, mu)
    return gsobject, safe

def _Magnify(gsobject, config, key, base, logger):
    mu, safe = ParseValue(config, key, base, float)
    logger.debug('obj %d: mu  = %f',base.get('obj_num',0),mu)
    gsobject = gsobject.magnify(mu)
    return gsobject, safe

def _Shift(gsobject, config, key, base, logger):
    shift, safe = ParseValue(config, key, base, PositionD)
    logger.debug('obj %d: shift  = %f,%f',base.get('obj_num',0),shift.x,shift.y)
    gsobject = gsobject._shift(shift.x, shift.y)
    return gsobject, safe

def RegisterObjectType(type_name, build_func, input_type=None):
    """Register an object type for use by the config apparatus.

    A few notes about the signature of the build functions:

    1. The config parameter is the dict for the current object to be generated.  So it should
       be the case that config['type'] == type_name.
    2. The base parameter is the original config dict being processed.
    3. The ignore parameter  is a list of items that should be ignored in the config dict if they
       are present and not valid for the object being built.
    4. The gsparams parameter is a dict of kwargs that should be used to build a GSParams object
       to use when building this object.
    5. The logger parameter is a logging.Logger object to use for logging progress if desired.
    6. The return value of build_func should be a tuple consisting of the object and a boolean,
       safe, which indicates whether the generated object is safe to use again rather than
       regenerate for subsequent postage stamps. e.g. if a PSF has all constant values, then it
       can be used for all the galaxies in a simulation, which lets it keep any FFTs that it has
       performed internally.  OpticalPSF is a good example of where this can have a significant
       speed up.

    Parameters:
        type_name:      The name of the 'type' specification in the config dict.
        build_func:     A function to build a GSObject from the config information.
                        The call signature is::

                            obj, safe = Build(config, base, ignore, gsparams, logger)

        input_type:     If the type requires an input object, give the key name of the input
                        type here.  (If it uses more than one, this may be a list.)
                        [default: None]
    """
    valid_gsobject_types[type_name] = build_func
    RegisterInputConnectedType(input_type, type_name)

RegisterObjectType('None', _BuildNone)
RegisterObjectType('Add', _BuildAdd)
RegisterObjectType('Sum', _BuildAdd)
RegisterObjectType('Convolve', _BuildConvolve)
RegisterObjectType('Convolution', _BuildConvolve)
RegisterObjectType('List', _BuildList)
RegisterObjectType('OpticalPSF', _BuildOpticalPSF)
RegisterObjectType('ChromaticOpticalPSF', _BuildChromaticOpticalPSF)
RegisterObjectType('ChromaticAtmosphere', _BuildChromaticAtmosphere)
RegisterObjectType('Eval', _BuildEval)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging
import numpy as np

from .image import ImageBuilder, FlattenNoiseVariance, RegisterImageType
from .value import ParseValue, GetAllParams
from .stamp import BuildStamps
from .noise import AddSky, AddNoise
from .input import ProcessInputNObjects
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..image import Image

# This file adds image type Scattered, which places individual stamps at arbitrary
# locations on a larger image.

class ScatteredImageBuilder(ImageBuilder):

    def setup(self, config, base, image_num, obj_num, ignore, logger):
        """Do the initialization and setup for building the image.

        This figures out the size that the image will be, but doesn't actually build it yet.

        Parameters:
            config:     The configuration dict for the image field.
            base:       The base configuration dict.
            image_num:  The current image number.
            obj_num:    The first object number in the image.
            ignore:     A list of parameters that are allowed to be in config that we can
                        ignore here. i.e. it won't be an error if these parameters are present.
            logger:     If given, a logger object to log progress.

        Returns:
            xsize, ysize
        """
        logger.debug('image %d: Building Scattered: image, obj = %d,%d',
                     image_num,image_num,obj_num)

        self.nobjects = self.getNObj(config, base, image_num, logger=logger)
        logger.debug('image %d: nobj = %d',image_num,self.nobjects)

        # These are allowed for Scattered, but we don't use them here.
        extra_ignore = [ 'image_pos', 'world_pos', 'stamp_size', 'stamp_xsize', 'stamp_ysize',
                         'nobjects' ]
        opt = { 'size' : int , 'xsize' : int , 'ysize' : int, 'dtype': None }
        params = GetAllParams(config, base, opt=opt, ignore=ignore+extra_ignore)[0]

        size = params.get('size',0)
        full_xsize = params.get('xsize',size)
        full_ysize = params.get('ysize',size)

        if (full_xsize <= 0) or (full_ysize <= 0):
            raise GalSimConfigError(
                "Both image.xsize and image.ysize need to be defined and > 0.")

        # If image_force_xsize and image_force_ysize were set in config, make sure it matches.
        if ( ('image_force_xsize' in base and full_xsize != base['image_force_xsize']) or
             ('image_force_ysize' in base and full_ysize != base['image_force_ysize']) ):
            raise GalSimConfigError(
                "Unable to reconcile required image xsize and ysize with provided "
                "xsize=%d, ysize=%d, "%(full_xsize,full_ysize))

        return full_xsize, full_ysize


    def buildImage(self, config, base, image_num, obj_num, logger):
        """Build an Image containing multiple objects placed at arbitrary locations.

        Parameters:
            config:     The configuration dict for the image field.
            base:       The base configuration dict.
            image_num:  The current image number.
            obj_num:    The first object number in the image.
            logger:     If given, a logger object to log progress.

        Returns:
            the final image and the current noise variance in the image as a tuple
        """
        full_image = base['current_image']

        if 'image_pos' in config and 'world_pos' in config:
            raise GalSimConfigValueError(
                "Both image_pos and world_pos specified for Scattered image.",
                (config['image_pos'], config['world_pos']))

        if ('image_pos' not in config and 'world_pos' not in config and
                not ('stamp' in base and
                    ('image_pos' in base['stamp'] or 'world_pos' in base['stamp']))):
            full_xsize = base['image_xsize']
            full_ysize = base['image_ysize']
            xmin = base['image_origin'].x
            xmax = xmin + full_xsize-1
            ymin = base['image_origin'].y
            ymax = ymin + full_ysize-1
            config['image_pos'] = {
                'type' : 'XY' ,
                'x' : { 'type' : 'Random' , 'min' : xmin , 'max' : xmax },
                'y' : { 'type' : 'Random' , 'min' : ymin , 'max' : ymax }
            }

        stamps, current_vars = BuildStamps(
                self.nobjects, base, logger=logger, obj_num=obj_num, do_noise=False)

        base['index_key'] = 'image_num'

        for k in range(self.nobjects):
            # This is our signal that the object was skipped.
            if stamps[k] is None: continue
            bounds = stamps[k].bounds & full_image.bounds
            logger.debug('image %d: full bounds = %s',image_num,str(full_image.bounds))
            logger.debug('image %d: stamp %d bounds = %s',image_num,k,str(stamps[k].bounds))
            logger.debug('image %d: Overlap = %s',image_num,str(bounds))
            if bounds.isDefined():
                full_image[bounds] += stamps[k][bounds]
            else:
                logger.info(
                    "Object centered at (%d,%d) is entirely off the main image, "
                    "whose bounds are (%d,%d,%d,%d)."%(
                        stamps[k].center.x, stamps[k].center.y,
                        full_image.bounds.xmin, full_image.bounds.xmax,
                        full_image.bounds.ymin, full_image.bounds.ymax))

        # Bring the image so far up to a flat noise variance
        current_var = FlattenNoiseVariance(
                base, full_image, stamps, current_vars, logger)

        return full_image, current_var

    def makeTasks(self, config, base, jobs, logger):
        """Turn a list of jobs into a list of tasks.

        Here we just have one job per task.

        Parameters:
            config:     The configuration dict for the image field.
            base:       The base configuration dict.
            jobs:       A list of jobs to split up into tasks.  Each job in the list is a
                        dict of parameters that includes 'image_num' and 'obj_num'.
            logger:     If given, a logger object to log progress.

        Returns:
            a list of tasks
        """
        return [ [ (job, k) ] for k, job in enumerate(jobs) ]

    def addNoise(self, image, config, base, image_num, obj_num, current_var, logger):
        """Add the final noise to a Scattered image

        Parameters:
            image:          The image onto which to add the noise.
            config:         The configuration dict for the image field.
            base:           The base configuration dict.
            image_num:      The current image number.
            obj_num:        The first object number in the image.
            current_var:    The current noise variance in each postage stamps.
            logger:         If given, a logger object to log progress.
        """
        base['current_noise_image'] = base['current_image']
        AddSky(base,image)
        AddNoise(base,image,current_var,logger)


    def getNObj(self, config, base, image_num, logger=None, approx=False):
        """Get the number of objects that will be built for this image.

        Parameters:
            config:         The configuration dict for the image field.
            base:           The base configuration dict.
            image_num:      The current image number.
            logger:         If given, a logger object to log progress.
            approx:         Whether an approximate/overestimate is ok [default: False]

        Returns:
            the number of objects
        """
        orig_index_key = base.get('index_key',None)
        base['index_key'] = 'image_num'
        base['image_num'] = image_num

        # Allow nobjects to be automatic based on input catalog
        if 'nobjects' not in config:
            nobj = ProcessInputNObjects(base, logger=logger, approx=approx)
            if nobj is None:
                raise GalSimConfigError(
                    "Attribute nobjects is required for image.type = Scattered")
        else:
            nobj = ParseValue(config,'nobjects',base,int)[0]
        base['index_key'] = orig_index_key
        return nobj

# Register this as a valid image type
RegisterImageType('Scattered', ScatteredImageBuilder())





# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from __future__ import print_function
import numpy as np
import os
import shutil
import sys
import logging

import galsim
from galsim_test_helpers import *

@timer
def test_input_init():
    """Test using the init method for inputs.
    """
    # Most of the tests in this file write to the 'output' directory.
    # Here we write to a different directory and make sure that it properly
    # creates the directory if necessary.
    if os.path.exists('output_fits'):
        shutil.rmtree('output_fits')
    config = {
        'modules': ['config_input_test_modules'],
        'input': {
            'input_size_module': [{'size': 55}, {'size': 45}],
        },
        'image': {
            'type': 'Single',
            'random_seed': 1234,
            'xsize': "$input_size_0",
            'ysize': "$input_size_arr[1]",
        },
        'gal': {
            'type': 'Gaussian',
            'sigma': {'type': 'Random', 'min': 1, 'max': 2},
            'flux': 100,
        },
        'output': {
            'type': 'Fits',
            'file_name': "output_fits/test_fits.fits"
        },
    }

    logger = logging.getLogger('test_fits')
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.DEBUG)

    first_seed = galsim.BaseDeviate(1234).raw()
    ud = galsim.UniformDeviate(first_seed + 0 + 1)
    sigma = ud() + 1.
    gal = galsim.Gaussian(sigma=sigma, flux=100)
    im1 = gal.drawImage(scale=1, nx=55, ny=45)

    galsim.config.Process(config, logger=logger)
    file_name = 'output_fits/test_fits.fits'
    im2 = galsim.fits.read(file_name)
    np.testing.assert_array_equal(im2.array, im1.array)

@timer
def test_approx_nobjects():
    """Test the getApproxNObjects functionality.
    """
    class BigCatalog(galsim.Catalog):
        def getApproxNObjects(self):
            return 2*self.getNObjects()

    class SmallCatalog(galsim.Catalog):
        def getApproxNObjects(self):
            return self.getNObjects() // 2

    def GenerateFromBigCatalog(config, base, value_type):
        input_cat = galsim.config.GetInputObj('big_catalog', config, base, 'BigCatalog')
        galsim.config.SetDefaultIndex(config, input_cat.getNObjects())
        req = {'col': int, 'index': int}
        kwargs, safe = galsim.config.GetAllParams(config, base, req=req)
        col = kwargs['col']
        index = kwargs['index']
        return input_cat.getFloat(index, col), safe

    def GenerateFromSmallCatalog(config, base, value_type):
        input_cat = galsim.config.GetInputObj('small_catalog', config, base, 'SmallCatalog')
        galsim.config.SetDefaultIndex(config, input_cat.getNObjects())
        req = {'col': int, 'index': int}
        kwargs, safe = galsim.config.GetAllParams(config, base, req=req)
        col = kwargs['col']
        index = kwargs['index']
        return input_cat.getFloat(index, col), safe

    galsim.config.RegisterInputType('big_catalog',
                                    galsim.config.InputLoader(BigCatalog, has_nobj=True))
    galsim.config.RegisterInputType('small_catalog',
                                    galsim.config.InputLoader(SmallCatalog, has_nobj=True))
    galsim.config.RegisterValueType('BigCatalog', GenerateFromBigCatalog, [float],
                                    input_type='big_catalog')
    galsim.config.RegisterValueType('SmallCatalog', GenerateFromSmallCatalog, [float],
                                    input_type='small_catalog')

    # FWIW, this is the config dict from test_variable_cat_size in test_config_image.py.
    config = {
        'gal': {
            'type': 'Gaussian',
            'half_light_radius': { 'type': 'Catalog', 'col': 0 },
            'shear': {
                'type': 'G1G2',
                'g1': { 'type': 'Catalog', 'col': 1 },
                'g2': { 'type': 'Catalog', 'col': 2 }
            },
            'flux': 1.7
        },
        'stamp': {
            'size': 33
        },
        'image': {
            'type': 'Scattered',
            'size': 256,
            'image_pos': {
                'type': 'XY',
                'x': { 'type': 'Catalog', 'col': 3 },
                'y': { 'type': 'Catalog', 'col': 4 }
            }
        },
        'input': {
            'catalog': {
                'dir': 'config_input',
                'file_name': [ 'cat_3.txt', 'cat_5.txt' ],
                'index_key': 'image_num',
            }
        },
        'output': {
            'type' : 'MultiFits',
            'file_name' : 'output/test_approx_nobj.fits',
            'nimages' : 2,
        }
    }
    config1 = galsim.config.CopyConfig(config)
    config2 = galsim.config.CopyConfig(config)
    config3 = galsim.config.CopyConfig(config)

    # First regular with normal Catalog.
    galsim.config.Process(config1)
    images1 = galsim.fits.readMulti('output/test_approx_nobj.fits')

    # Now with catalogs whose approximate nobj are too big.
    config2['gal']['half_light_radius']['type'] = 'BigCatalog'
    config2['gal']['shear']['g1']['type'] = 'BigCatalog'
    config2['gal']['shear']['g2']['type'] = 'BigCatalog'
    config2['image']['image_pos']['x']['type'] = 'BigCatalog'
    config2['image']['image_pos']['y']['type'] = 'BigCatalog'
    config2['input']['big_catalog'] = galsim.config.CopyConfig(config['input']['catalog'])
    del config2['input']['catalog']
    galsim.config.Process(config2)
    images2 = galsim.fits.readMulti('output/test_approx_nobj.fits')
    assert images2[0] == images1[0]
    assert images2[1] == images1[1]

    # Lastly, one where the approx nobj is too small.
    config3['gal']['half_light_radius']['type'] = 'SmallCatalog'
    config3['gal']['shear']['g1']['type'] = 'SmallCatalog'
    config3['gal']['shear']['g2']['type'] = 'SmallCatalog'
    config3['image']['image_pos']['x']['type'] = 'SmallCatalog'
    config3['image']['image_pos']['y']['type'] = 'SmallCatalog'
    config3['input']['small_catalog'] = galsim.config.CopyConfig(config['input']['catalog'])
    del config3['input']['catalog']
    with CaptureLog() as cl:
        galsim.config.Process(config3, logger=cl.logger)
    # The real numbers are 3 and 5, but the small version guesses them to be 1 and 2 respectively.
    assert 'Input small_catalog has approximately 1' in cl.output
    assert 'Input small_catalog has approximately 2' in cl.output

    images3 = galsim.fits.readMulti('output/test_approx_nobj.fits')
    assert images3[0] == images1[0]
    assert images3[1] == images1[1]


@timer
def test_atm_input():
    """Test an imsim-like AtmosphericPSF as an input object.
    """
    import multiprocessing

    class AtmPSF:

        def __init__(self):
            ctx = multiprocessing.get_context('fork')
            rng = galsim.BaseDeviate(1234)
            atm = galsim.Atmosphere(mp_context=ctx, r0_500=0.2,
                                    altitude=[0.2, 2.58, 5.16],  # Same as imsim, but just 3 layers
                                    r0_weights=[0.652, 0.172, 0.055],
                                    rng=rng, screen_size=256, screen_scale=1)
            self.aper = galsim.Aperture(diam=2, lam=500, screen_list=atm)
            with galsim.utilities.single_threaded():
                with ctx.Pool(3, initializer=galsim.phase_screens.initWorker,
                              initargs=galsim.phase_screens.initWorkerArgs()) as pool:
                    atm.instantiate(pool=pool, check='FFT')
            self.atm = atm

        def getGSScreenShare(self):
            return self.atm[0].getGSScreenShare()

        def getPSF(self):
            return self.atm.makePSF(500, aper=self.aper)

    def BuildAtmPSF(config, base, ignore, gsparams, logger):
        atm = galsim.config.GetInputObj('atm_psf', config, base, 'AtmPSF')
        return atm.getPSF(), True

    galsim.config.RegisterInputType('atm_psf',
            galsim.config.InputLoader(AtmPSF, use_proxy=False,
                                      worker_init=galsim.phase_screens.initWorker,
                                      worker_initargs=galsim.phase_screens.initWorkerArgs))
    galsim.config.RegisterObjectType('AtmPSF', BuildAtmPSF, input_type='atm_psf')

    config = {
        'input': {
            'atm_psf': {}
        },
        'image': {
            'type': 'Single',
            'random_seed': 1234,
        },
        'gal': {
            'type': 'Gaussian',
            'sigma': {'type': 'Random', 'min': 1, 'max': 2},
            'flux': 100,
        },
        'psf': {
            'type': 'AtmPSF',
        },
        'output': {
            'type': 'MultiFits',
            'nimages': 3,
            'file_name': "output_fits/test_atm.fits",
        },
    }

    # Basically the check here is that multiprocessing works here
    # and that it produces the same images as serial processing

    # First nproc=1.
    config1 = galsim.config.CopyConfig(config)
    galsim.config.Process(config1)

    # Now nproc=3 in output field.
    # This doesn't actually do 3 processes, since there is only 1 output file,
    # but even this fails without the use_proxy=False bit above.
    config2 = galsim.config.CopyConfig(config)
    config2['output']['nproc'] = 3
    config2['output']['file_name'] = "output_fits/test_atm_output_nproc_3.fits"
    galsim.config.Process(config2)

    # Finally nproc=3 in image field.
    # This really does parallelize the run over the 3 images.
    config3 = galsim.config.CopyConfig(config)
    config3['image']['nproc'] = 3
    config3['output']['file_name'] = "output_fits/test_atm_image_nproc_3.fits"
    galsim.config.Process(config3)

    # Check that the results match
    ims1 = galsim.fits.readMulti("output_fits/test_atm.fits")
    ims2 = galsim.fits.readMulti("output_fits/test_atm_output_nproc_3.fits")
    ims3 = galsim.fits.readMulti("output_fits/test_atm_image_nproc_3.fits")

    for im1, im2, im3 in zip(ims1,ims2,ims3):
        assert im1 == im2
        assert im1 == im3

    # Both worker_init and worker_initargs are required when either is provided.
    with assert_raises(galsim.GalSimError):
        galsim.config.InputLoader(AtmPSF, use_proxy=False,
                                  worker_init=galsim.phase_screens.initWorker)
    with assert_raises(galsim.GalSimError):
        galsim.config.InputLoader(AtmPSF, use_proxy=False,
                                  worker_initargs=galsim.phase_screens.initWorkerArgs)

@timer
def test_dependent_inputs():
    """Test inputs that depend on other inputs.

    imSim has input types that depend on other input types.  If these are listed in order,
    with dependencies first, everything is fine.  But when using recursive templates, this can
    be difficult to get right.  So we now have a check in GalSim when loading in input object,
    if it raises an exception that indicates it needs a different input type, try to load that
    one first.
    """
    class Dict1:
        def __init__(self):
            self.d = {'a': 1, 'b': 2}

    def Dict1Item(config, base, value_type):
        d = galsim.config.GetInputObj('dict1', config, base, 'Dict1').d
        key, safe = galsim.config.ParseValue(config, 'key', base, str)
        return d[key], safe

    class Dict2:
        def __init__(self):
            self.d = {'c': 1, 'd': 2}

    def Dict2Item(config, base, value_type):
        d = galsim.config.GetInputObj('dict2', config, base, 'Dict2').d
        key, safe = galsim.config.ParseValue(config, 'key', base, str)
        return d[key], safe

    class Dep:
        _req_params = dict(a=float, b=float, c=float, d=float)

        def __init__(self, a, b, c, d):
            self.d = dict(a=a, b=b, c=c, d=d)

    def DepItem(config, base, value_type):
        d = galsim.config.GetInputObj('dep', config, base, 'Dep').d
        key, safe = galsim.config.ParseValue(config, 'key', base, str)
        return d[key], safe

    galsim.config.RegisterInputType('dict1', galsim.config.InputLoader(Dict1))
    galsim.config.RegisterValueType('Dict1Item', Dict1Item, input_type='dict1', valid_types=[float])
    galsim.config.RegisterInputType('dict2', galsim.config.InputLoader(Dict2))
    galsim.config.RegisterValueType('Dict2Item', Dict2Item, input_type='dict2', valid_types=[float])
    galsim.config.RegisterInputType('dep', galsim.config.InputLoader(Dep))
    galsim.config.RegisterValueType('DepItem', DepItem, input_type='dep', valid_types=[float])

    # First put the input items in order, so all dependencies are resolved before they are needed.
    config = {
        'input': {
            'dict1': {},
            'dict2': {},
            'dep': {
                'a': {'type': 'Dict1Item', 'key': 'a'},
                'b': {'type': 'Dict1Item', 'key': 'b'},
                'c': {'type': 'Dict2Item', 'key': 'c'},
                'd': {'type': 'Dict2Item', 'key': 'd'},
            },
        }
    }
    with CaptureLog() as cl:
        galsim.config.ProcessInput(config, cl.logger)
    assert 'input seems to depend on' not in cl.output
    dep = galsim.config.GetInputObj('dep', config, config, 'Dep')
    assert dep.d == dict(a=1, b=2, c=1, d=2)

    # Now put them out of order.
    config = {
        'input': {
            'dep': {
                'a': {'type': 'Dict1Item', 'key': 'a'},
                'b': {'type': 'Dict1Item', 'key': 'b'},
                'c': {'type': 'Dict2Item', 'key': 'c'},
                'd': {'type': 'Dict2Item', 'key': 'd'},
            },
            'dict1': {},
            'dict2': {},
        }
    }
    with CaptureLog() as cl:
        galsim.config.ProcessInput(config, cl.logger)
    assert 'dep input seems to depend on dict1' in cl.output
    assert 'dep input seems to depend on dict2' in cl.output
    dep = galsim.config.GetInputObj('dep', config, config, 'Dep')
    assert dep.d == dict(a=1, b=2, c=1, d=2)

    # If it depends on something that in turn fails, it doesn't work.
    class Dict3:
        _opt_params = dict(e=float)
        _takes_rng = True
        def __init__(self, rng, e=None):
            ud = galsim.UniformDeviate(rng)
            self.d = dict(a=ud(), b=ud(), c=ud(), d=ud())
            if e: self.d['e'] = e

    def Dict3Item(config, base, value_type):
        d = galsim.config.GetInputObj('dict3', config, base, 'Dict3').d
        key, safe = galsim.config.ParseValue(config, 'key', base, str)
        return d[key], safe

    galsim.config.RegisterInputType('dict3', galsim.config.InputLoader(Dict3))
    galsim.config.RegisterValueType('Dict3Item', Dict3Item, input_type='dict3', valid_types=[float])

    config = {
        'input': {
            'dep': {
                'a': {'type': 'Dict1Item', 'key': 'a'},
                'b': {'type': 'Dict3Item', 'key': 'b'},
                'c': {'type': 'Dict2Item', 'key': 'c'},
                'd': {'type': 'Dict3Item', 'key': 'd'},
            },
            'dict1': {},
            'dict2': {},
            'dict3': {
                'e': {'type': 'Dict4Item', 'key': 'a'},
            }
        }
    }
    with np.testing.assert_raises(galsim.config.GalSimConfigError):
        galsim.config.ProcessInput(config, cl.logger)

    # But with safe_only=True, it doesn't raise an exception
    config = galsim.config.CleanConfig(config)
    with CaptureLog() as cl:
        galsim.config.ProcessInput(config, cl.logger, safe_only=True)

    assert 'dep input seems to depend on dict1' in cl.output
    assert 'dep input seems to depend on dict2' not in cl.output  # Doesn't get to the Dict2Item
    assert 'dep input seems to depend on dict3' in cl.output
    dep = galsim.config.GetInputObj('dep', config, config, 'Dep')
    assert dep is None

    # If the dependency graph is circular, make sure we don't get an infinite loop.
    config = galsim.config.CleanConfig(config)
    config['input']['dict3']['e'] = {'type': 'DepItem', 'key': 'a'}
    galsim.config.ProcessInput(config, cl.logger, safe_only=True)
    dep = galsim.config.GetInputObj('dep', config, config, 'Dep')
    assert dep is None

    # Finally, just make sure that if dict3 isn't broken, it all works as expected.
    config = galsim.config.CleanConfig(config)
    config['input']['dict3']['e']['type'] = 'Dict1Item'
    galsim.config.ProcessInput(config, cl.logger)
    dict3 = galsim.config.GetInputObj('dict3', config, config, 'Dict3')
    dep = galsim.config.GetInputObj('dep', config, config, 'Dep')
    print('dict3.d = ',dict3.d)
    print('dep.d = ',dep.d)
    assert dep.d == dict(a=1, b=dict3.d['b'], c=1, d=dict3.d['d'])


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging

from .input import InputLoader, GetInputObj, RegisterInputType
from .gsobject import RegisterObjectType
from .util import GetRNG
from .value import GetAllParams, SetDefaultIndex, RegisterValueType
from ..errors import GalSimConfigError
from ..gsparams import GSParams
from ..galaxy_sample import GalaxySample, COSMOSCatalog

# This file adds input types cosmos_catalog and sample_galaxy and gsobject types
# COSMOSGalaxy and SampleGalaxy.

# The SampleGalaxy doesn't need anything special other than registration as a valid input type.
# However, we do make a custom Loader so that we can add a logger line with some information about
# the number of objects in the catalog that passed the initial cuts and other basic catalog info.
class SampleLoader(InputLoader):
    def __init__(self, cls, input_field):
        self.cls_name = cls.__name__
        self.input_field = input_field
        super().__init__(cls)

    def setupImage(self, cosmos_cat, config, base, logger):
        if logger:
            # Only report as a warning the first time.  After that, use info.
            first = not base.get('_SampleLoader_reported_as_warning',False)
            base['_SampleLoader_reported_as_warning'] = True
            if first:
                log_level = logging.WARNING
            else:
                log_level = logging.INFO
            # It should be required that base['input']['cosmos_catalog'] exists, but
            # just in case someone calls this in a weird way, use get() with defaults.
            cc = base.get('input',{}).get(self.input_field,{})
            if isinstance(cc,list): cc = cc[0]
            out_str = ''
            if 'sample' in cc:
                out_str += '\n  sample = %s'%cc['sample']
            if 'dir' in cc:
                out_str += '\n  dir = %s'%cc['dir']
            if 'file_name' in cc:
                out_str += '\n  file_name = %s'%cc['file_name']
            if out_str != '':
                logger.log(log_level, 'Using user-specified %s: %s',self.cls_name, out_str)
            logger.info("file %d: Sample catalog has %d total objects; %d passed initial cuts.",
                        base['file_num'], cosmos_cat.getNTot(), cosmos_cat.nobjects)
            if base.get('gal',{}).get('gal_type',None) == 'parametric':
                logger.log(log_level,"Using parametric galaxies.")
            else:
                logger.log(log_level,"Using real galaxies.")

RegisterInputType('cosmos_catalog', SampleLoader(COSMOSCatalog, 'cosmos_catalog'))
RegisterInputType('galaxy_sample', SampleLoader(GalaxySample, 'galaxy_sample'))

# The gsobject types coupled to these are COSMOSGalaxy and SampleGalaxy respectively.

def _BuildCOSMOSGalaxy(config, base, ignore, gsparams, logger):
    """Build a COSMOS galaxy using the cosmos_catalog input item.
    """
    sample_cat = GetInputObj('cosmos_catalog', config, base, 'COSMOSGalaxy')
    return _FinishBuildSampleGalaxy(config, base, ignore, gsparams, logger,
                                    sample_cat, 'COSMOSGalaxy')

def _BuildSampleGalaxy(config, base, ignore, gsparams, logger):
    """Build a sample galaxy using the galaxy_sample input item.
    """
    sample_cat = GetInputObj('galaxy_sample', config, base, 'SampleGalaxy')
    return _FinishBuildSampleGalaxy(config, base, ignore, gsparams, logger,
                                    sample_cat, 'SampleGalaxy')

def _FinishBuildSampleGalaxy(config, base, ignore, gsparams, logger, sample_cat, cls_name):
    ignore = ignore + ['num']

    # Special: if galaxies are selected based on index, and index is Sequence or Random, and max
    # isn't set, set it to nobjects-1.
    if 'index' in config:
        SetDefaultIndex(config, sample_cat.nobjects)

    opt = { "index" : int,
            "gal_type" : str,
            "noise_pad_size" : float,
            "deep" : bool,
            "sersic_prec": float,
            "chromatic": bool,
            "area": float,
            "exptime": float,
    }

    kwargs, safe = GetAllParams(config, base, opt=opt, ignore=ignore)
    if gsparams: kwargs['gsparams'] = GSParams(**gsparams)

    rng = GetRNG(config, base, logger, cls_name)

    if 'index' not in kwargs:
        kwargs['index'], n_rng_calls = sample_cat.selectRandomIndex(1, rng=rng, _n_rng_calls=True)
        safe = False

        # Make sure this process gives consistent results regardless of the number of processes
        # being used.
        if not isinstance(sample_cat, GalaxySample) and rng is not None:
            # Then sample_cat is really a proxy, which means the rng was pickled, so we need to
            # discard the same number of random calls from the one in the config dict.
            rng.discard(int(n_rng_calls))

    kwargs['rng'] = rng

    # NB. Even though index is officially optional, it will always be present, either because it was
    #     set by a call to selectRandomIndex, explicitly by the user, or due to the call to
    #     SetDefaultIndex.
    index = kwargs['index']
    if index >= sample_cat.nobjects:
        raise GalSimConfigError(
            "index=%s has gone past the number of entries in the %s"%(index, cls_name))

    logger.debug('obj %d: %s kwargs = %s',base.get('obj_num',0), cls_name, kwargs)

    kwargs['self'] = sample_cat

    # Use a staticmethod of GalaxySample to avoid pickling the result of makeGalaxy()
    # The RealGalaxy in particular has a large serialization, so it is more efficient to
    # make it in this process, which is what happens here.
    gal = GalaxySample._makeGalaxy(**kwargs)

    return gal, safe

# Register this as a valid gsobject type
RegisterObjectType('COSMOSGalaxy', _BuildCOSMOSGalaxy, input_type='cosmos_catalog')
RegisterObjectType('SampleGalaxy', _BuildSampleGalaxy, input_type='galaxy_sample')

# Finally, also provide accessor to values in the parametric catalog

def _GetCOSMOSValue(config, base, value_type):
    cosmos_cat = GetInputObj('cosmos_catalog', config, base, 'COSMOSValue')
    req = { 'key': str, 'index': int }
    kwargs, safe = GetAllParams(config, base, req=req)
    return cosmos_cat.getValue(**kwargs)

def _GetSampleValue(config, base, value_type):
    sample_cat = GetInputObj('galaxy_sample', config, base, 'SampleValue')
    req = { 'key': str, 'index': int }
    kwargs, safe = GetAllParams(config, base, req=req)
    return sample_cat.getValue(**kwargs)

RegisterValueType('COSMOSValue', _GetCOSMOSValue, [float], input_type='cosmos_catalog')
RegisterValueType('SampleValue', _GetSampleValue, [float], input_type='galaxy_sample')
